# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module for Filecoin address encoding/decoding."""

# Imports
from enum import IntEnum, unique
from typing import Any, Union

from bip_utils.addr.addr_dec_utils import AddrDecUtils
from bip_utils.addr.addr_key_validator import AddrKeyValidator
from bip_utils.addr.iaddr_decoder import IAddrDecoder
from bip_utils.addr.iaddr_encoder import IAddrEncoder
from bip_utils.coin_conf import CoinsConf
from bip_utils.ecc import IPublicKey
from bip_utils.utils.crypto import Blake2b32, Blake2b160
from bip_utils.utils.misc import Base32Decoder, Base32Encoder, IntegerUtils


@unique
class FillAddrTypes(IntEnum):
    """Enumerative for Filecoin address types."""

    SECP256K1 = 1
    BLS = 3


class FilAddrConst:
    """Class container for Filecoin address constants."""

    # Alphabet for base32
    BASE32_ALPHABET: str = "abcdefghijklmnopqrstuvwxyz234567"


class _FilAddrUtils:
    """Class container for Filecoin address utility functions."""

    @staticmethod
    def ComputeChecksum(pub_key_hash: bytes,
                        addr_type: FillAddrTypes) -> bytes:
        """
        Compute checksum in EOS format.

        Args:
            pub_key_hash (bytes)     : Public key hash
            addr_type (FillAddrTypes): Address type

        Returns:
            bytes: Computed checksum
        """
        addr_type_byte = IntegerUtils.ToBytes(addr_type)
        return Blake2b32.QuickDigest(addr_type_byte + pub_key_hash)

    @staticmethod
    def DecodeAddr(addr: str,
                   addr_type: FillAddrTypes) -> bytes:
        """
        Decode a Filecoin address to bytes.

        Args:
            addr (str)               : Address string
            addr_type (FillAddrTypes): Address type

        Returns:
            bytes: Public key hash bytes

        Raises:
            ValueError: If the address encoding is not valid
        """

        # Validate and remove prefix
        addr_no_prefix = AddrDecUtils.ValidateAndRemovePrefix(addr, CoinsConf.Filecoin.ParamByKey("addr_prefix"))
        # Check address type
        addr_type_got = ord(addr_no_prefix[0]) - ord("0")
        if addr_type != addr_type_got:
            raise ValueError(f"Invalid address type (expected {addr_type}, got {addr_type_got})")
        # Decode from base32
        addr_dec_bytes = Base32Decoder.Decode(addr_no_prefix[1:], FilAddrConst.BASE32_ALPHABET)
        # Validate length
        AddrDecUtils.ValidateLength(addr_dec_bytes,
                                    Blake2b160.DigestSize() + Blake2b32.DigestSize())

        # Get back checksum and public key bytes
        pub_key_hash_bytes, checksum_bytes = AddrDecUtils.SplitPartsByChecksum(addr_dec_bytes,
                                                                               Blake2b32.DigestSize())
        # Validate checksum
        AddrDecUtils.ValidateChecksum(pub_key_hash_bytes, checksum_bytes,
                                      lambda pub_key_bytes: _FilAddrUtils.ComputeChecksum(pub_key_bytes, addr_type))

        return pub_key_hash_bytes

    @staticmethod
    def EncodeKeyBytes(pub_key_bytes: bytes,
                       addr_type: FillAddrTypes) -> str:
        """
        Encode a public key to Filecoin address.

        Args:
            pub_key_bytes (bytes)    : Public key bytes
            addr_type (FillAddrTypes): Address type

        Returns:
            str: Address string
        """

        # Get address type
        addr_type_str = chr(addr_type + ord("0"))

        # Compute public key hash and checksum
        pub_key_hash_bytes = Blake2b160.QuickDigest(pub_key_bytes)
        checksum_bytes = _FilAddrUtils.ComputeChecksum(pub_key_hash_bytes, addr_type)
        # Encode to base32
        b32_enc = Base32Encoder.EncodeNoPadding(pub_key_hash_bytes + checksum_bytes, FilAddrConst.BASE32_ALPHABET)

        return CoinsConf.Filecoin.ParamByKey("addr_prefix") + addr_type_str + b32_enc


class FilSecp256k1AddrDecoder(IAddrDecoder):
    """
    Filecoin address decoder class, based on secp256k1 curve.
    It allows the Filecoin address decoding.
    """

    @staticmethod
    def DecodeAddr(addr: str,
                   **kwargs: Any) -> bytes:
        """
        Decode a Filecoin address to bytes.

        Args:
            addr (str): Address string
            **kwargs  : Not used

        Returns:
            bytes: Public key hash bytes

        Raises:
            ValueError: If the address encoding is not valid
        """
        return _FilAddrUtils.DecodeAddr(addr, FillAddrTypes.SECP256K1)


class FilSecp256k1AddrEncoder(IAddrEncoder):
    """
    Filecoin address encoder class, based on secp256k1 curve.
    It allows the Filecoin address encoding.
    """

    @staticmethod
    def EncodeKey(pub_key: Union[bytes, IPublicKey],
                  **kwargs: Any) -> str:
        """
        Encode a public key to Filecoin address.

        Args:
            pub_key (bytes or IPublicKey): Public key bytes or object
            **kwargs                     : Not used

        Returns:
            str: Address string

        Raised:
            ValueError: If the public key is not valid
            TypeError: If the public key is not secp256k1 or the address type is not valid
        """
        pub_key_obj = AddrKeyValidator.ValidateAndGetSecp256k1Key(pub_key)
        pub_key_bytes = pub_key_obj.RawUncompressed().ToBytes()

        return _FilAddrUtils.EncodeKeyBytes(pub_key_bytes, FillAddrTypes.SECP256K1)


# Deprecated: only for compatibility, Encoder class shall be used instead
FilSecp256k1Addr = FilSecp256k1AddrEncoder



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module containing common classes for mnemonic."""

# Imports
from __future__ import annotations

from enum import Enum
from typing import List, Union


class MnemonicLanguages(Enum):
    """Base enum for mnemonic languages."""


class Mnemonic:
    """
    Mnemonic class. It represents a generic mnemonic phrase.
    It acts as a simple container with some helper functions, so it doesn't validate the given mnemonic.
    """

    m_mnemonic_list: List[str]

    @classmethod
    def FromString(cls,
                   mnemonic_str: str) -> Mnemonic:
        """
        Create a class from mnemonic string.

        Args:
            mnemonic_str (str): Mnemonic string

        Returns:
            Mnemonic: Mnemonic object
        """
        return cls.FromList(cls._Normalize(mnemonic_str))

    @classmethod
    def FromList(cls,
                 mnemonic_list: List[str]) -> Mnemonic:
        """
        Create a class from mnemonic list.

        Args:
            mnemonic_list (list[str]): Mnemonic list

        Returns:
            Mnemonic: Mnemonic object
        """
        return cls(cls._Normalize(mnemonic_list))

    def __init__(self,
                 mnemonic_list: List[str]) -> None:
        """
        Construct class.

        Args:
            mnemonic_list (list[str]): Mnemonic list
        """
        self.m_mnemonic_list = mnemonic_list

    def WordsCount(self) -> int:
        """
        Get the words count.

        Returns:
            int: Words count
        """
        return len(self.m_mnemonic_list)

    def ToList(self) -> List[str]:
        """
        Get the mnemonic as a list.

        Returns:
            list[str]: Mnemonic as a list
        """
        return self.m_mnemonic_list

    def ToStr(self) -> str:
        """
        Get the mnemonic as a string.

        Returns:
            str: Mnemonic as a string
        """
        return " ".join(self.m_mnemonic_list)

    def __str__(self) -> str:
        """
        Get the mnemonic as a string.

        Returns:
            str: Mnemonic as a string
        """
        return self.ToStr()

    @staticmethod
    def _Normalize(mnemonic: Union[str, List[str]]) -> List[str]:
        """
        Normalize mnemonic list.

        Args:
            mnemonic (str or list[str]): Mnemonic

        Returns:
            list[str]: Normalized mnemonic list
        """
        if isinstance(mnemonic, str):
            mnemonic = mnemonic.split()
        return mnemonic



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module with ed25519 constants."""

from bip_utils.ecc.common.ipoint import IPoint

# Imports
from bip_utils.ecc.ed25519.ed25519_point import Ed25519Point


class Ed25519Const:
    """Class container for Ed25519 constants."""

    # Curve name
    NAME: str = "Ed25519"
    # Curve order
    CURVE_ORDER: int = 2**252 + 27742317777372353535851937790883648493
    # Curve generator point
    GENERATOR: IPoint = Ed25519Point.FromCoordinates(
        15112221349535400772501151409588531511454012693041857206046113283949847762202,
        46316835694926478169428394003475163141307993866256225615783033603165251855960
    )



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module for Algorand mnemonic utility classes."""

# Imports
from typing import List, Optional, Union

from bip_utils.algorand.mnemonic.algorand_mnemonic import AlgorandMnemonicConst
from bip_utils.utils.crypto import Sha512_256


class AlgorandMnemonicUtils:
    """Class container for Algorand mnemonic utility functions."""

    @staticmethod
    def ComputeChecksum(data_bytes: bytes) -> bytes:
        """
        Compute checksum.

        Args:
            data_bytes (bytes): Data bytes

        Returns:
            bytes: Computed checksum
        """
        return Sha512_256.QuickDigest(data_bytes)[:AlgorandMnemonicConst.CHECKSUM_BYTE_LEN]

    @staticmethod
    def ComputeChecksumWordIndex(data_bytes: bytes) -> int:
        """
        Compute checksum word index.

        Args:
            data_bytes (bytes): Data bytes

        Returns:
            str: Computed checksum word index
        """

        # Compute checksum and convert it to 11-bit
        chksum = AlgorandMnemonicUtils.ComputeChecksum(data_bytes)
        chksum_11bit = AlgorandMnemonicUtils.ConvertBits(chksum, 8, 11)
        # Cannot be None by converting bytes from 8-bit to 11-bit
        assert chksum_11bit is not None

        return chksum_11bit[0]

    @staticmethod
    def ConvertBits(data: Union[bytes, List[int]],
                    from_bits: int,
                    to_bits: int) -> Optional[List[int]]:
        """
        Perform bit conversion.
        The function takes the input data (list of integers or byte sequence) and convert every value from
        the specified number of bits to the specified one.
        It returns a list of integer where every number is less than 2^to_bits.

        Args:
            data (list[int] or bytes): Data to be converted
            from_bits (int)          : Number of bits to start from
            to_bits (int)            : Number of bits to end with

        Returns:
            list[int]: List of converted values, None in case of errors
        """
        max_out_val = (1 << to_bits) - 1

        acc = 0
        bits = 0
        ret = []

        for value in data:
            # Value shall not be less than zero or greater than 2^from_bits
            if value < 0 or (value >> from_bits):
                return None
            # Continue accumulating until greater than to_bits
            acc |= value << bits
            bits += from_bits
            while bits >= to_bits:
                ret.append(acc & max_out_val)
                acc = acc >> to_bits
                bits -= to_bits

        if bits != 0:
            ret.append(acc & max_out_val)

        return ret



# Copyright (c) 2022 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module for Electrum v1 mnemonic generation."""

# Imports
from typing import Dict, Union

from bip_utils.electrum.mnemonic_v1.electrum_v1_entropy_generator import (
    ElectrumV1EntropyBitLen, ElectrumV1EntropyGenerator
)
from bip_utils.electrum.mnemonic_v1.electrum_v1_mnemonic import (
    ElectrumV1Languages, ElectrumV1MnemonicConst, ElectrumV1WordsNum
)
from bip_utils.electrum.mnemonic_v1.electrum_v1_mnemonic_encoder import ElectrumV1MnemonicEncoder
from bip_utils.utils.mnemonic import Mnemonic


class ElectrumV1MnemonicGeneratorConst:
    """Class container for Electrum v1 mnemonic generator constants."""

    # Entropy length for each words number
    WORDS_NUM_TO_ENTROPY_LEN: Dict[ElectrumV1WordsNum, ElectrumV1EntropyBitLen] = {
        ElectrumV1WordsNum.WORDS_NUM_12: ElectrumV1EntropyBitLen.BIT_LEN_128,
    }


class ElectrumV1MnemonicGenerator:
    """
    Electrum v1 mnemonic generator class.
    It generates 12-words mnemonic in according to v1 Electrum mnemonic.
    """

    m_mnemonic_encoder: ElectrumV1MnemonicEncoder

    def __init__(self,
                 lang: ElectrumV1Languages = ElectrumV1Languages.ENGLISH) -> None:
        """
        Construct class.

        Args:
            lang (ElectrumV1Languages, optional): Language (default: English)

        Raises:
            TypeError: If the language is not a ElectrumV1Languages enum
            ValueError: If language words list is not valid
        """
        self.m_mnemonic_encoder = ElectrumV1MnemonicEncoder(lang)

    def FromWordsNumber(self,
                        words_num: Union[int, ElectrumV1WordsNum]) -> Mnemonic:
        """
        Generate mnemonic with the specified words number from random entropy.
        There is no really need of this method, since the words number can only be 12, but it's
        kept to have the same usage of Bip39/Monero mnemonic generator.

        Args:
            words_num (int or ElectrumV1WordsNum): Number of words (12)

        Returns:
            Mnemonic object: Generated mnemonic

        Raises:
            ValueError: If words number is not valid
        """

        # Check words number
        if words_num not in ElectrumV1MnemonicConst.MNEMONIC_WORD_NUM:
            raise ValueError(f"Words number for mnemonic ({words_num}) is not valid")

        # Convert int to enum if necessary
        if isinstance(words_num, int):
            words_num = ElectrumV1WordsNum(words_num)

        # Get entropy length in bit from words number
        entropy_bit_len = ElectrumV1MnemonicGeneratorConst.WORDS_NUM_TO_ENTROPY_LEN[words_num]
        # Generate entropy
        entropy_bytes = ElectrumV1EntropyGenerator(entropy_bit_len).Generate()

        return self.FromEntropy(entropy_bytes)

    def FromEntropy(self,
                    entropy_bytes: bytes) -> Mnemonic:
        """
        Generate mnemonic from the specified entropy bytes.

        Args:
            entropy_bytes (bytes): Entropy bytes

        Returns:
            Mnemonic object: Generated mnemonic

        Raises:
            ValueError: If entropy byte length is not valid
        """
        return self.m_mnemonic_encoder.Encode(entropy_bytes)



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module for ed25519-monero keys."""

# Imports
from nacl import signing

from bip_utils.ecc.common.ikeys import IPrivateKey, IPublicKey
from bip_utils.ecc.common.ipoint import IPoint
from bip_utils.ecc.curve.elliptic_curve_types import EllipticCurveTypes
from bip_utils.ecc.ed25519.ed25519_keys import Ed25519KeysConst, Ed25519PrivateKey, Ed25519PublicKey
from bip_utils.ecc.ed25519.lib import ed25519_lib
from bip_utils.ecc.ed25519_monero.ed25519_monero_point import Ed25519MoneroPoint
from bip_utils.utils.misc import DataBytes


class Ed25519MoneroPublicKey(Ed25519PublicKey):
    """Ed25519-Monero public key class."""

    @staticmethod
    def CurveType() -> EllipticCurveTypes:
        """
        Get the elliptic curve type.

        Returns:
           EllipticCurveTypes: Elliptic curve type
        """
        return EllipticCurveTypes.ED25519_MONERO

    @staticmethod
    def CompressedLength() -> int:
        """
        Get the compressed key length.

        Returns:
           int: Compressed key length
        """
        return Ed25519KeysConst.PUB_KEY_BYTE_LEN

    @staticmethod
    def UncompressedLength() -> int:
        """
        Get the uncompressed key length.

        Returns:
           int: Uncompressed key length
        """
        return Ed25519MoneroPublicKey.CompressedLength()

    def RawCompressed(self) -> DataBytes:
        """
        Return raw compressed public key.

        Returns:
            DataBytes object: DataBytes object
        """
        return DataBytes(bytes(self.m_ver_key))

    def Point(self) -> IPoint:
        """
        Get public key point.

        Returns:
            IPoint object: IPoint object
        """
        return Ed25519MoneroPoint(bytes(self.m_ver_key))


class Ed25519MoneroPrivateKey(Ed25519PrivateKey):
    """Ed25519-Monero private key class."""

    @classmethod
    def FromBytes(cls,
                  key_bytes: bytes) -> IPrivateKey:
        """
        Construct class from key bytes.

        Args:
            key_bytes (bytes): Key bytes

        Returns:
            IPrivateKey: IPrivateKey object

        Raises:
            ValueError: If key bytes are not valid
        """
        if not ed25519_lib.scalar_is_valid(key_bytes):
            raise ValueError("Invalid private key bytes")
        return super().FromBytes(key_bytes)

    @staticmethod
    def CurveType() -> EllipticCurveTypes:
        """
        Get the elliptic curve type.

        Returns:
           EllipticCurveTypes: Elliptic curve type
        """
        return EllipticCurveTypes.ED25519_MONERO

    def PublicKey(self) -> IPublicKey:
        """
        Get the public key correspondent to the private one.

        Returns:
            IPublicKey object: IPublicKey object
        """
        return Ed25519MoneroPublicKey(
            signing.VerifyKey(
                ed25519_lib.point_scalar_mul_base(bytes(self.m_sign_key))
            )
        )



from bip_utils.monero.monero import Monero
from bip_utils.monero.monero_ex import MoneroKeyError
from bip_utils.monero.monero_keys import MoneroPrivateKey, MoneroPublicKey
from bip_utils.monero.monero_subaddr import MoneroSubaddress



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module for getting Substrate coins configuration."""

# Imports
from typing import Dict

from bip_utils.substrate.conf.substrate_coin_conf import SubstrateCoinConf
from bip_utils.substrate.conf.substrate_coins import SubstrateCoins
from bip_utils.substrate.conf.substrate_conf import SubstrateConf


class SubstrateConfGetterConst:
    """Class container for Substrate configuration getter constants."""

    # Map from SubstrateCoins to configuration classes
    COIN_TO_CONF: Dict[SubstrateCoins, SubstrateCoinConf] = {
        SubstrateCoins.ACALA: SubstrateConf.Acala,
        SubstrateCoins.BIFROST: SubstrateConf.Bifrost,
        SubstrateCoins.CHAINX: SubstrateConf.ChainX,
        SubstrateCoins.EDGEWARE: SubstrateConf.Edgeware,
        SubstrateCoins.GENERIC: SubstrateConf.Generic,
        SubstrateCoins.KARURA: SubstrateConf.Karura,
        SubstrateCoins.KUSAMA: SubstrateConf.Kusama,
        SubstrateCoins.MOONBEAM: SubstrateConf.Moonbeam,
        SubstrateCoins.MOONRIVER: SubstrateConf.Moonriver,
        SubstrateCoins.PHALA: SubstrateConf.Phala,
        SubstrateCoins.PLASM: SubstrateConf.Plasm,
        SubstrateCoins.POLKADOT: SubstrateConf.Polkadot,
        SubstrateCoins.SORA: SubstrateConf.Sora,
        SubstrateCoins.STAFI: SubstrateConf.Stafi,
    }


class SubstrateConfGetter:
    """
    Substrate configuration getter class.
    It allows to get the Substrate configuration of a specific coin.
    """

    @staticmethod
    def GetConfig(coin_type: SubstrateCoins) -> SubstrateCoinConf:
        """
        Get coin configuration.

        Args:
            coin_type (SubstrateCoins): Coin type

        Returns:
            SubstrateCoinConf: Coin configuration

        Raises:
            TypeError: If coin type is not of a SubstrateCoins enumerative
        """
        if not isinstance(coin_type, SubstrateCoins):
            raise TypeError("Coin type is not an enumerative of SubstrateCoins")
        return SubstrateConfGetterConst.COIN_TO_CONF[coin_type]



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Module for BIP39 mnemonic decoding.
Reference: https://github.com/bitcoin/bips/blob/master/bip-0039.mediawiki
"""

# Imports
from typing import Optional, Union

from bip_utils.bip.bip39.bip39_mnemonic import Bip39Languages, Bip39Mnemonic, Bip39MnemonicConst
from bip_utils.bip.bip39.bip39_mnemonic_utils import Bip39WordsListFinder, Bip39WordsListGetter
from bip_utils.utils.crypto import Sha256
from bip_utils.utils.misc import BytesUtils, IntegerUtils
from bip_utils.utils.mnemonic import Mnemonic, MnemonicChecksumError, MnemonicDecoderBase, MnemonicWordsList


class Bip39MnemonicDecoder(MnemonicDecoderBase):
    """
    BIP39 mnemonic decoder class.
    It decodes a mnemonic phrase to bytes.
    """

    def __init__(self,
                 lang: Optional[Bip39Languages] = None) -> None:
        """
        Construct class.

        Args:
            lang (Bip39Languages, optional): Language, None for automatic detection

        Raises:
            TypeError: If the language is not a Bip39Languages enum
            ValueError: If loaded words list is not valid
        """
        super().__init__(lang, Bip39WordsListFinder, Bip39WordsListGetter)

    def Decode(self,
               mnemonic: Union[str, Mnemonic]) -> bytes:
        """
        Decode a mnemonic phrase to bytes (no checksum).

        Args:
            mnemonic (str or Mnemonic object): Mnemonic

        Returns:
            bytes: Decoded bytes (no checksum)

        Raises:
            MnemonicChecksumError: If checksum is not valid
            ValueError: If mnemonic is not valid
        """
        mnemonic_bin_str = self.__DecodeAndVerifyBinaryStr(mnemonic)

        return self.__EntropyBytesFromBinaryStr(mnemonic_bin_str)

    def DecodeWithChecksum(self,
                           mnemonic: Union[str, Mnemonic]) -> bytes:
        """
        Decode a mnemonic phrase to bytes (with checksum).

        Args:
            mnemonic (str or Mnemonic object): Mnemonic

        Returns:
            bytes: Decoded bytes (with checksum)

        Raises:
            MnemonicChecksumError: If checksum is not valid
            ValueError: If mnemonic is not valid
        """
        mnemonic_bin_str = self.__DecodeAndVerifyBinaryStr(mnemonic)

        # Compute pad bit length
        mnemonic_bit_len = len(mnemonic_bin_str)
        pad_bit_len = (mnemonic_bit_len
                       if mnemonic_bit_len % 8 == 0
                       else mnemonic_bit_len + (8 - mnemonic_bit_len % 8))

        return BytesUtils.FromBinaryStr(mnemonic_bin_str, pad_bit_len // 4)

    def __DecodeAndVerifyBinaryStr(self,
                                   mnemonic: Union[str, Mnemonic]) -> str:
        """
        Decode a mnemonic phrase to its mnemonic binary string by verifying the checksum.

        Args:
            mnemonic (str or Mnemonic object): Mnemonic

        Returns:
            str: Mnemonic binary string

        Raises:
            MnemonicChecksumError: If checksum is not valid
            ValueError: If mnemonic is not valid
        """
        mnemonic_obj = Bip39Mnemonic.FromString(mnemonic) if isinstance(mnemonic, str) else mnemonic

        # Check mnemonic length
        if mnemonic_obj.WordsCount() not in Bip39MnemonicConst.MNEMONIC_WORD_NUM:
            raise ValueError(f"Mnemonic words count is not valid ({mnemonic_obj.WordsCount()})")

        # Detect language if it was not specified at construction
        words_list, _ = self._FindLanguage(mnemonic_obj)

        # Get back mnemonic binary string
        mnemonic_bin_str = self.__MnemonicToBinaryStr(mnemonic_obj, words_list)

        # Verify checksum
        checksum_bin_str = mnemonic_bin_str[-self.__GetChecksumLen(mnemonic_bin_str):]
        checksum_bin_str_got = self.__ComputeChecksumBinaryStr(mnemonic_bin_str)

        if checksum_bin_str != checksum_bin_str_got:
            raise MnemonicChecksumError(
                f"Invalid checksum (expected {checksum_bin_str}, got {checksum_bin_str_got})"
            )

        return mnemonic_bin_str

    def __ComputeChecksumBinaryStr(self,
                                   mnemonic_bin_str: str) -> str:
        """
        Compute checksum from mnemonic binary string.

        Args:
            mnemonic_bin_str (str): Mnemonic binary string

        Returns:
           str: Computed checksum binary string
        """

        # Get entropy bytes
        entropy_bytes = self.__EntropyBytesFromBinaryStr(mnemonic_bin_str)
        # Convert entropy hash to binary string
        entropy_hash_bin_str = BytesUtils.ToBinaryStr(Sha256.QuickDigest(entropy_bytes),
                                                      Sha256.DigestSize() * 8)

        # Return checksum
        return entropy_hash_bin_str[:self.__GetChecksumLen(mnemonic_bin_str)]

    def __EntropyBytesFromBinaryStr(self,
                                    mnemonic_bin_str: str) -> bytes:
        """
        Get entropy bytes from mnemonic binary string.

        Args:
            mnemonic_bin_str (str): Mnemonic binary string

        Returns:
           bytes: Entropy bytes
        """

        # Get checksum length
        checksum_len = self.__GetChecksumLen(mnemonic_bin_str)
        # Get back entropy binary string
        entropy_bin_str = mnemonic_bin_str[:-checksum_len]

        # Get entropy bytes from binary string
        return BytesUtils.FromBinaryStr(entropy_bin_str, checksum_len * 8)

    @staticmethod
    def __MnemonicToBinaryStr(mnemonic: Mnemonic,
                              words_list: MnemonicWordsList) -> str:
        """
        Get mnemonic binary string from mnemonic phrase.

        Args:
            mnemonic (Mnemonic object)           : Mnemonic object
            words_list (MnemonicWordsList object): Words list object

        Returns:
           str: Mnemonic binary string

        Raises:
            ValueError: If the one of the mnemonic word is not valid
        """

        # Convert each word to its index in binary format
        mnemonic_bin_str = map(lambda word: IntegerUtils.ToBinaryStr(words_list.GetWordIdx(word),
                                                                     Bip39MnemonicConst.WORD_BIT_LEN),
                               mnemonic.ToList())

        return "".join(mnemonic_bin_str)

    @staticmethod
    def __GetChecksumLen(mnemonic_bin_str: str) -> int:
        """
        Get checksum length from mnemonic binary string.

        Args:
            mnemonic_bin_str (str): Mnemonic binary string

        Returns:
           int: Checksum length
        """
        return len(mnemonic_bin_str) // 33



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Module for BIP38 encryption/decryption.
Reference: https://github.com/bitcoin/bips/blob/master/bip-0038.mediawiki
"""

# Imports
from typing import Optional, Tuple, Union

from bip_utils.bip.bip38.bip38_addr import Bip38PubKeyModes
from bip_utils.bip.bip38.bip38_ec import Bip38EcDecrypter, Bip38EcKeysGenerator
from bip_utils.bip.bip38.bip38_no_ec import Bip38NoEcDecrypter, Bip38NoEcEncrypter
from bip_utils.ecc import IPrivateKey


class Bip38Encrypter:
    """
    BIP38 encrypter class.
    It encrypts a private key using the algorithm specified in BIP38.
    """

    @staticmethod
    def EncryptNoEc(priv_key: Union[bytes, IPrivateKey],
                    passphrase: str,
                    pub_key_mode: Bip38PubKeyModes = Bip38PubKeyModes.COMPRESSED) -> str:
        """
        Encrypt the specified private key without EC multiplication.

        Args:
            priv_key (bytes or IPrivateKey)          : Private key bytes or object
            passphrase (str)                         : Passphrase
            pub_key_mode (Bip38PubKeyModes, optional): Public key mode

        Returns:
            str: Encrypted private key

        Raises:
            TypeError: If the private key is not a Secp256k1PrivateKey
            ValueError: If the private key bytes are not valid
        """
        return Bip38NoEcEncrypter.Encrypt(priv_key, passphrase, pub_key_mode)

    @staticmethod
    def GeneratePrivateKeyEc(passphrase: str,
                             pub_key_mode: Bip38PubKeyModes = Bip38PubKeyModes.COMPRESSED,
                             lot_num: Optional[int] = None,
                             sequence_num: Optional[int] = None) -> str:
        """
        Generate a random encrypted private key with EC multiplication, using the specified parameters.
        This will generate the intermediate passphrase and use it immediately for generating the private key.

        Args:
            passphrase (str)                         : Passphrase
            pub_key_mode (Bip38PubKeyModes, optional): Public key mode
            lot_num (int, optional)                  : Lot number
            sequence_num (int, optional)             : Sequence number

        Returns:
            str: Encrypted private key
        """
        int_pass = Bip38EcKeysGenerator.GenerateIntermediatePassphrase(passphrase,
                                                                       lot_num,
                                                                       sequence_num)
        return Bip38EcKeysGenerator.GeneratePrivateKey(int_pass, pub_key_mode)


class Bip38Decrypter:
    """
    BIP38 decrypter class.
    It decrypts a private key using the algorithm specified in BIP38.
    """

    @staticmethod
    def DecryptNoEc(priv_key_enc: str,
                    passphrase: str) -> Tuple[bytes, Bip38PubKeyModes]:
        """
        Decrypt the specified private key without EC multiplication.

        Args:
            priv_key_enc (str): Encrypted private key bytes
            passphrase (str)  : Passphrase

        Returns:
            tuple[bytes, Bip38PubKeyModes]: Decrypted private key (index 0), public key mode (index 1)

        Raises:
            Base58ChecksumError: If base58 checksum is not valid
            ValueError: If the encrypted key is not valid
        """
        return Bip38NoEcDecrypter.Decrypt(priv_key_enc, passphrase)

    @staticmethod
    def DecryptEc(priv_key_enc: str,
                  passphrase: str) -> Tuple[bytes, Bip38PubKeyModes]:
        """
        Decrypt the specified private key with EC multiplication.

        Args:
            priv_key_enc (str): Encrypted private key bytes
            passphrase (str)  : Passphrase

        Returns:
            tuple[bytes, Bip38PubKeyModes]: Decrypted private key (index 0), public key mode (index 1)

        Raises:
            Base58ChecksumError: If base58 checksum is not valid
            ValueError: If the encrypted key is not valid
        """
        return Bip38EcDecrypter.Decrypt(priv_key_enc, passphrase)



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module for Monero mnemonic encoding."""

# Imports
from abc import ABC
from typing import List

from bip_utils.monero.mnemonic.monero_entropy_generator import MoneroEntropyGenerator
from bip_utils.monero.mnemonic.monero_mnemonic import MoneroLanguages, MoneroMnemonic
from bip_utils.monero.mnemonic.monero_mnemonic_utils import MoneroMnemonicUtils, MoneroWordsListGetter
from bip_utils.utils.mnemonic import Mnemonic, MnemonicEncoderBase, MnemonicUtils


class MoneroMnemonicEncoderBase(MnemonicEncoderBase, ABC):
    """
    Monero mnemonic encoder base class.
    It encodes bytes to the mnemonic phrase.
    """

    m_lang: MoneroLanguages

    def __init__(self,
                 lang: MoneroLanguages = MoneroLanguages.ENGLISH) -> None:
        """
        Construct class.

        Args:
            lang (MoneroLanguages, optional): Language (default: English)

        Raises:
            TypeError: If the language is not a Bip39Languages enum
            ValueError: If loaded words list is not valid
        """
        super().__init__(lang, MoneroWordsListGetter)
        self.m_lang = lang

    def _EncodeToList(self,
                      entropy_bytes: bytes) -> List[str]:
        """
        Encode bytes to list of mnemonic words.

        Args:
            entropy_bytes (bytes): Entropy bytes (accepted lengths in bits: 128, 256)

        Returns:
            list[str]: List of encoded mnemonic words

        Raises:
            ValueError: If bytes length is not valid
        """

        # Check entropy length
        entropy_byte_len = len(entropy_bytes)
        if not MoneroEntropyGenerator.IsValidEntropyByteLen(entropy_byte_len):
            raise ValueError(f"Entropy byte length ({entropy_byte_len}) is not valid")

        # Consider 4 bytes at a time, 4 bytes represent 3 words
        mnemonic = []
        for i in range(len(entropy_bytes) // 4):
            mnemonic += MnemonicUtils.BytesChunkToWords(entropy_bytes[i * 4:(i * 4) + 4], self.m_words_list, "little")

        return mnemonic


class MoneroMnemonicNoChecksumEncoder(MoneroMnemonicEncoderBase):
    """
    Monero mnemonic encoder class (no checksum).
    It encodes bytes to the mnemonic phrase without checksum.
    """

    def Encode(self,
               entropy_bytes: bytes) -> Mnemonic:
        """
        Encode bytes to mnemonic phrase (no checksum).

        Args:
            entropy_bytes (bytes): Entropy bytes (accepted lengths in bits: 128, 256)

        Returns:
            Mnemonic object: Encoded mnemonic (no checksum)

        Raises:
            ValueError: If entropy is not valid
        """
        return MoneroMnemonic.FromList(self._EncodeToList(entropy_bytes))


class MoneroMnemonicWithChecksumEncoder(MoneroMnemonicEncoderBase):
    """
    Monero mnemonic encoder class (with checksum).
    It encodes bytes to the mnemonic phrase with checksum.
    """

    def Encode(self,
               entropy_bytes: bytes) -> Mnemonic:
        """
        Encode bytes to mnemonic phrase (with checksum).

        Args:
            entropy_bytes (bytes): Entropy bytes (accepted lengths in bits: 128, 256)

        Returns:
            Mnemonic object: Encoded mnemonic (with checksum)

        Raises:
            ValueError: If entropy is not valid
        """
        words = self._EncodeToList(entropy_bytes)
        checksum_word = MoneroMnemonicUtils.ComputeChecksum(words, self.m_lang)

        return MoneroMnemonic.FromList(words + [checksum_word])


class MoneroMnemonicEncoder:
    """
    Monero mnemonic encoder class.
    Helper class to encode bytes to the mnemonic phrase with or without checksum.
    """

    m_no_chk_enc: MoneroMnemonicNoChecksumEncoder
    m_with_chk_enc: MoneroMnemonicWithChecksumEncoder

    def __init__(self,
                 lang: MoneroLanguages = MoneroLanguages.ENGLISH) -> None:
        """
        Construct class.

        Args:
            lang (MoneroLanguages, optional): Language (default: English)

        Raises:
            TypeError: If the language is not a MoneroLanguages enum
            ValueError: If loaded words list is not valid
        """
        self.m_no_chk_enc = MoneroMnemonicNoChecksumEncoder(lang)
        self.m_with_chk_enc = MoneroMnemonicWithChecksumEncoder(lang)

    def EncodeNoChecksum(self,
                         entropy_bytes: bytes) -> Mnemonic:
        """
        Encode bytes to mnemonic phrase (no checksum).

        Args:
            entropy_bytes (bytes): Entropy bytes (accepted lengths in bits: 128, 256)

        Returns:
            Mnemonic object: Encoded mnemonic (no checksum)

        Raises:
            ValueError: If bytes length is not valid
        """
        return self.m_no_chk_enc.Encode(entropy_bytes)

    def EncodeWithChecksum(self,
                           entropy_bytes: bytes) -> Mnemonic:
        """
        Encode bytes to mnemonic phrase (with checksum).

        Args:
            entropy_bytes (bytes): Entropy bytes (accepted lengths in bits: 128, 256)

        Returns:
            Mnemonic object: Encoded mnemonic (with checksum)

        Raises:
            ValueError: If bytes length is not valid
        """
        return self.m_with_chk_enc.Encode(entropy_bytes)



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module for Substrate keys computation and derivation."""

# Imports
from __future__ import annotations

from typing import Optional, Union

import sr25519

from bip_utils.ecc import IPrivateKey, IPublicKey
from bip_utils.substrate.conf import SubstrateCoinConf, SubstrateCoins, SubstrateConfGetter
from bip_utils.substrate.substrate_ex import SubstrateKeyError
from bip_utils.substrate.substrate_keys import SubstratePrivateKey, SubstratePublicKey
from bip_utils.substrate.substrate_path import SubstratePath, SubstratePathElem, SubstratePathParser


class SubstrateConst:
    """Class container for Substrate constants."""

    # Seed minimum length in bytes
    SEED_MIN_BYTE_LEN: int = 32


class Substrate:
    """
    Substrate class.
    It allows to compute Substrate keys and addresses.
    """

    m_priv_key: Optional[SubstratePrivateKey]
    m_pub_key: SubstratePublicKey
    m_path: SubstratePath
    m_coin_conf: SubstrateCoinConf

    #
    # Construction methods
    #

    @classmethod
    def FromSeed(cls,
                 seed_bytes: bytes,
                 coin_type: SubstrateCoins) -> Substrate:
        """
        Create a Substrate object from the specified seed.

        Args:
            seed_bytes (bytes)        : Seed bytes
            coin_type (SubstrateCoins): Coin type

        Returns:
            Substrate object: Substrate object

        Raises:
            TypeError: If coin_type is not of SubstrateCoins enum
            ValueError: If the seed length is not valid
        """
        if len(seed_bytes) < SubstrateConst.SEED_MIN_BYTE_LEN:
            raise ValueError(
                f"Seed length is too small, it shall be at least {SubstrateConst.SEED_MIN_BYTE_LEN} bytes"
            )

        pub_key_bytes, priv_key_bytes = sr25519.pair_from_seed(  # pylint: disable=no-member
            seed_bytes[:SubstrateConst.SEED_MIN_BYTE_LEN]
        )
        return cls(priv_key=priv_key_bytes,
                   pub_key=pub_key_bytes,
                   path=SubstratePath(),
                   coin_conf=SubstrateConfGetter.GetConfig(coin_type))

    @classmethod
    def FromSeedAndPath(cls,
                        seed_bytes: bytes,
                        path: Union[str, SubstratePath],
                        coin_type: SubstrateCoins) -> Substrate:
        """
        Create a Substrate object from the specified seed and path.

        Args:
            seed_bytes (bytes)                : Seed bytes
            path (str or SubstratePath object): Path
            coin_type (SubstrateCoins)        : Coin type

        Returns:
            Substrate object: Substrate object

        Raises:
            TypeError: If coin_type is not of SubstrateCoins enum
            ValueError: If the seed length is not valid
            SubstratePathError: If the path is not valid
        """
        substrate_ctx = cls.FromSeed(seed_bytes, coin_type)
        return substrate_ctx.DerivePath(path)

    @classmethod
    def FromPrivateKey(cls,
                       priv_key: Union[bytes, IPrivateKey],
                       coin_type: SubstrateCoins) -> Substrate:
        """
        Create a Substrate object from the specified private key.

        Args:
            priv_key (bytes or IPrivateKey): Private key
            coin_type (SubstrateCoins)     : Coin type

        Returns:
            Substrate object: Substrate object

        Raises:
            TypeError: If coin_type is not of SubstrateCoins enum
            SubstrateKeyError: If the key is not valid
        """
        return cls(priv_key=priv_key,
                   pub_key=None,
                   path=SubstratePath(),
                   coin_conf=SubstrateConfGetter.GetConfig(coin_type))

    @classmethod
    def FromPublicKey(cls,
                      pub_key: Union[bytes, IPublicKey],
                      coin_type: SubstrateCoins) -> Substrate:
        """
        Create a Substrate object from the specified public key.

        Args:
            pub_key (bytes or IPublicKey): Public key
            coin_type (SubstrateCoins)   : Coin type

        Returns:
            Substrate object: Substrate object

        Raises:
            TypeError: If coin_type is not of SubstrateCoins enum
            SubstrateKeyError: If the key is not valid
        """
        if not isinstance(coin_type, SubstrateCoins):
            raise TypeError("Coin is not an enumerative of SubstrateCoins")

        return cls(priv_key=None,
                   pub_key=pub_key,
                   path=SubstratePath(),
                   coin_conf=SubstrateConfGetter.GetConfig(coin_type))

    #
    # Public methods
    #

    def __init__(self,
                 priv_key: Optional[Union[bytes, IPrivateKey]],
                 pub_key: Optional[Union[bytes, IPublicKey]],
                 path: SubstratePath,
                 coin_conf: SubstrateCoinConf) -> None:
        """
        Construct class.

        Args:
            priv_key (bytes or IPrivateKey)     : Private key, if None a public-only object will be created
            pub_key (bytes or IPublicKey)       : Public key
            path (SubstratePath object)         : Path
            coin_conf (SubstrateCoinConf object): SubstrateCoinConf object

        Raises:
            SubstrateKeyError: If one of the key is not valid
        """

        # Private key object
        if priv_key is not None:
            self.m_priv_key = SubstratePrivateKey.FromBytesOrKeyObject(priv_key, coin_conf)
            # Use the provided public key if any. This is done because se25519 library returns both the
            # derived private and public keys, so we can avoid to compute the public key from the private key
            # that takes time
            self.m_pub_key = (SubstratePublicKey.FromBytesOrKeyObject(pub_key, coin_conf)
                              if pub_key is not None
                              else self.m_priv_key.PublicKey())
        # Public-only object
        else:
            assert isinstance(pub_key, (bytes, IPublicKey)), "Public key shall be specified for public-only objects"

            self.m_priv_key = None
            self.m_pub_key = SubstratePublicKey.FromBytesOrKeyObject(pub_key, coin_conf)

        self.m_path = path
        self.m_coin_conf = coin_conf

    def ChildKey(self,
                 path_elem: Union[str, SubstratePathElem]) -> Substrate:
        """
        Create and return a child key of the current one with the specified path element.

        Args:
            path_elem (str or SubstratePathElem object): Path element

        Returns:
            Substrate object: Substrate object

        Raises:
            SubstrateKeyError: If the index results in invalid keys
        """
        if isinstance(path_elem, str):
            path_elem = SubstratePathElem(path_elem)

        return self.__CkdPriv(path_elem) if not self.IsPublicOnly() else self.__CkdPub(path_elem)

    def DerivePath(self,
                   path: Union[str, SubstratePath]) -> Substrate:
        """
        Derive children keys from the specified path.

        Args:
            path (str or SubstratePath object): Path

        Returns:
            Substrate object: Substrate object

        Raises:
            SubstratePathError: If the path is not valid
        """
        if isinstance(path, str):
            path = SubstratePathParser.Parse(path)

        substrate_obj = self
        # Derive children keys
        for path_elem in path:
            substrate_obj = substrate_obj.ChildKey(path_elem)

        return substrate_obj

    def ConvertToPublic(self) -> None:
        """Convert a private Substrate object into a public one."""
        self.m_priv_key = None

    def IsPublicOnly(self) -> bool:
        """
        Get if it's public-only.

        Returns:
            bool: True if public-only, false otherwise
        """
        return self.m_priv_key is None

    def CoinConf(self) -> SubstrateCoinConf:
        """
        Return coin configuration.

        Returns:
            SubstrateCoinConf object: SubstrateCoinConf object
        """
        return self.m_coin_conf

    def Path(self) -> SubstratePath:
        """
        Return path.

        Returns:
            SubstratePath object: SubstratePath object
        """
        return self.m_path

    def PrivateKey(self) -> SubstratePrivateKey:
        """
        Return private key object.

        Returns:
            SubstratePrivateKey object: SubstratePrivateKey object

        Raises:
            SubstrateKeyError: If internal key is public-only
        """
        if self.IsPublicOnly():
            raise SubstrateKeyError("Public-only deterministic keys have no private half")

        assert isinstance(self.m_priv_key, SubstratePrivateKey)
        return self.m_priv_key

    def PublicKey(self) -> SubstratePublicKey:
        """
        Return public key object.

        Returns:
            SubstratePublicKey object: SubstratePublicKey object
        """
        return self.m_pub_key

    #
    # Private methods
    #

    def __CkdPriv(self,
                  path_elem: SubstratePathElem) -> Substrate:
        """
        Create a child key of the specified path element using private derivation.

        Args:
            path_elem (SubstratePathElem object): Path element

        Returns:
            Substrate object: Substrate object
        """
        assert isinstance(self.m_priv_key, SubstratePrivateKey)

        ex_key_pair = (path_elem.ChainCode(), self.m_pub_key.RawCompressed().ToBytes(), self.m_priv_key.Raw().ToBytes())

        if path_elem.IsHard():
            _, pub_key_bytes, priv_key_bytes = sr25519.hard_derive_keypair(ex_key_pair,  # pylint: disable=no-member
                                                                           b"")
        else:
            _, pub_key_bytes, priv_key_bytes = sr25519.derive_keypair(ex_key_pair,  # pylint: disable=no-member
                                                                      b"")

        return Substrate(priv_key=priv_key_bytes,
                         pub_key=pub_key_bytes,
                         path=self.m_path.AddElem(path_elem),
                         coin_conf=self.m_coin_conf)

    def __CkdPub(self,
                 path_elem: SubstratePathElem) -> Substrate:
        """
        Create a child key of the specified index using public derivation.

        Args:
            path_elem (SubstratePathElem object): Path element

        Returns:
            Substrate object: Substrate object

        Raises:
            SubstrateKeyError: If the path element is hard
        """
        if path_elem.IsHard():
            raise SubstrateKeyError("Public child derivation cannot be used to create a hardened child key")

        ex_key_pair = (path_elem.ChainCode(), self.m_pub_key.RawCompressed().ToBytes())

        _, pub_key_bytes = sr25519.derive_pubkey(ex_key_pair, b"")  # pylint: disable=no-member

        return Substrate(priv_key=None,
                         pub_key=pub_key_bytes,
                         path=self.m_path.AddElem(path_elem),
                         coin_conf=self.m_coin_conf)



# Copyright (c) 2022 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Module for Cardano Shelley keys derivation.
Reference: https://cips.cardano.org/cips/cip11
"""

# Imports
from __future__ import annotations

import copy
from functools import lru_cache

from bip_utils.addr import AdaShelleyStakingAddrEncoder
from bip_utils.bip.bip44_base import Bip44Base, Bip44Changes, Bip44Levels
from bip_utils.cardano.cip1852 import Cip1852
from bip_utils.cardano.shelley.cardano_shelley_keys import CardanoShelleyPrivateKeys, CardanoShelleyPublicKeys


class CardanoShelley:
    """
    Cardano Shelley class.
    It allows keys derivation and addresses computation (including the staking one) in according to Cardano Shelley.
    """

    m_bip_obj: Bip44Base
    m_bip_sk_obj: Bip44Base

    @classmethod
    def FromCip1852Object(cls,
                          bip_obj: Bip44Base) -> CardanoShelley:
        """
        Create a CardanoShelley object from the specified Cip1852 object.

        Args:
            bip_obj (Bip44Base object): Bip44Base object

        Returns:
            CardanoShelley object: CardanoShelley object

        Raises:
            ValueError: If the seed is too short
            Bip32KeyError: If the seed is not suitable for master key generation
        """
        if not isinstance(bip_obj, Cip1852):
            raise ValueError("The Bip object shall be a Cip1852 instance")
        return cls(bip_obj,
                   cls.__DeriveStakingKeys(bip_obj))

    def __init__(self,
                 bip_obj: Bip44Base,
                 bip_sk_obj: Bip44Base) -> None:
        """
        Construct class.

        Args:
            bip_obj (Bip44Base object)   : Bip44Base object
            bip_sk_obj (Bip44Base object): Bip44Base object (staking)

        Raises:
            ValueError: If the bip_sk_obj object is below account level
        """
        if bip_obj.Level() < Bip44Levels.ACCOUNT:
            raise ValueError("The bip_obj shall not be below account level")
        if bip_sk_obj.Level() != Bip44Levels.ADDRESS_INDEX:
            raise ValueError("The bip_sk_obj shall be of address index level")
        self.m_bip_obj = bip_obj
        self.m_bip_sk_obj = bip_sk_obj

    @lru_cache()
    def PublicKeys(self) -> CardanoShelleyPublicKeys:
        """
        Return the public keys.

        Returns:
            CardanoShelleyPublicKeys object: CardanoShelleyPublicKeys object
        """
        return CardanoShelleyPublicKeys(self.m_bip_obj.PublicKey().Bip32Key(),
                                        self.m_bip_sk_obj.PublicKey().Bip32Key(),
                                        self.m_bip_obj.CoinConf())

    @lru_cache()
    def PrivateKeys(self) -> CardanoShelleyPrivateKeys:
        """
        Return the private keys.

        Returns:
            CardanoShelleyPrivateKeys object: CardanoShelleyPrivateKeys object

        Raises:
            Bip32KeyError: If the Bip32 object is public-only
        """
        return CardanoShelleyPrivateKeys(self.m_bip_obj.PrivateKey().Bip32Key(),
                                         self.m_bip_sk_obj.PrivateKey().Bip32Key(),
                                         self.m_bip_obj.CoinConf())

    def RewardObject(self) -> Bip44Base:
        """
        Alias for StakingObject.

        Returns:
            Bip44Base object: Bip44Base object
        """
        return self.StakingObject()

    def StakingObject(self) -> Bip44Base:
        """
        Return the staking object.

        Returns:
            Bip44Base object: Bip44Base object
        """
        return self.m_bip_sk_obj

    def IsPublicOnly(self) -> bool:
        """
        Get if it's public-only.

        Returns:
            bool: True if public-only, false otherwise
        """
        return self.m_bip_obj.IsPublicOnly()

    def Change(self,
               change_type: Bip44Changes) -> CardanoShelley:
        """
        Derive a child key from the specified change type and return a new CardanoShelley object.

        Args:
            change_type (Bip44Changes): Change type, must a Bip44Changes enum

        Returns:
            CardanoShelley object: CardanoShelley object

        Raises:
            TypeError: If change type is not a Bip44Changes enum
            Bip44DepthError: If current depth is not suitable for deriving keys
            Bip32KeyError: If the derivation results in an invalid key
        """
        return CardanoShelley(self.m_bip_obj.Change(change_type),
                              self.m_bip_sk_obj)

    def AddressIndex(self,
                     addr_idx: int) -> CardanoShelley:
        """
        Derive a child key from the specified address index and return a new CardanoShelley object.

        Args:
            addr_idx (int): Address index

        Returns:
            CardanoShelley object: CardanoShelley object

        Raises:
            Cip1852DepthError: If current depth is not suitable for deriving keys
            Bip32KeyError: If the derivation results in an invalid key
        """
        return CardanoShelley(self.m_bip_obj.AddressIndex(addr_idx),
                              self.m_bip_sk_obj)

    @staticmethod
    def __DeriveStakingKeys(bip_obj: Bip44Base) -> Bip44Base:
        """
        Derive staking keys from a Bip44Base object.

        Args:
            bip_obj (Bip44Base object): Bip44Base object

        Returns:
            Bip44Base object: Bip44Base object

        Raises:
            Bip32KeyError: If the derivation results in an invalid key
        """

        # Create a new configuration with the staking address class (no need to deep-copying)
        coin_conf = copy.copy(bip_obj.CoinConf())
        coin_conf.m_addr_cls = AdaShelleyStakingAddrEncoder
        # Create Cip1852 object for staking keys
        return Cip1852(bip_obj.Bip32Object().DerivePath("2/0"),
                       coin_conf)



# Copyright (c) 2022 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module for Electrum v2 mnemonic."""

# Imports
from enum import Enum, IntEnum, auto, unique
from typing import Dict, List

from bip_utils.bip.bip39 import Bip39Languages, Bip39Mnemonic
from bip_utils.bip.bip39.bip39_mnemonic import Bip39MnemonicConst
from bip_utils.utils.mnemonic import MnemonicLanguages


@unique
class ElectrumV2WordsNum(IntEnum):
    """Enumerative for Electrum words number (v2)."""

    WORDS_NUM_12 = 12
    WORDS_NUM_24 = 24


@unique
class ElectrumV2Languages(MnemonicLanguages):
    """Enumerative for Electrum languages (v2)."""

    CHINESE_SIMPLIFIED = Bip39Languages.CHINESE_SIMPLIFIED
    ENGLISH = Bip39Languages.ENGLISH
    PORTUGUESE = Bip39Languages.PORTUGUESE
    SPANISH = Bip39Languages.SPANISH


@unique
class ElectrumV2MnemonicTypes(Enum):
    """Enumerative for Electrum v2 mnemonic types."""

    STANDARD = auto()       # Standard wallet
    SEGWIT = auto()         # Segwit wallet
    STANDARD_2FA = auto()   # Standard 2FA wallet
    SEGWIT_2FA = auto()     # Segwit 2FA wallet


class ElectrumV2MnemonicConst:
    """Class container for Electrum v2 mnemonic constants."""

    # Accepted mnemonic word numbers
    MNEMONIC_WORD_NUM: List[ElectrumV2WordsNum] = [
        ElectrumV2WordsNum.WORDS_NUM_12,
        ElectrumV2WordsNum.WORDS_NUM_24,
    ]

    # Mnemonic types to prefix
    TYPE_TO_PREFIX: Dict[ElectrumV2MnemonicTypes, str] = {
        ElectrumV2MnemonicTypes.STANDARD: "01",
        ElectrumV2MnemonicTypes.SEGWIT: "100",
        ElectrumV2MnemonicTypes.STANDARD_2FA: "101",
        ElectrumV2MnemonicTypes.SEGWIT_2FA: "102",
    }

    # Word length in bit
    WORD_BIT_LEN: int = Bip39MnemonicConst.WORD_BIT_LEN


class ElectrumV2Mnemonic(Bip39Mnemonic):
    """Electrum mnemonic class."""



# Copyright (c) 2022 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module for Electrum v1 mnemonic entropy generation."""

# Imports
from enum import IntEnum, unique
from typing import List, Union

from bip_utils.utils.mnemonic import EntropyGenerator


@unique
class ElectrumV1EntropyBitLen(IntEnum):
    """Enumerative for Electrum entropy bit lengths (v1)."""

    BIT_LEN_128 = 128


class ElectrumV1EntropyGeneratorConst:
    """Class container for Electrum entropy generator constants (v1)."""

    # Accepted entropy lengths in bit
    ENTROPY_BIT_LEN: List[ElectrumV1EntropyBitLen] = [
        ElectrumV1EntropyBitLen.BIT_LEN_128,
    ]


class ElectrumV1EntropyGenerator(EntropyGenerator):
    """
    Electrum entropy generator class (v1).
    It generates random entropy bytes.
    """

    def __init__(self,
                 bit_len: Union[int, ElectrumV1EntropyBitLen] = ElectrumV1EntropyBitLen.BIT_LEN_128) -> None:
        """
        Construct class.

        Args:
            bit_len (int or ElectrumV1EntropyBitLen, optional): Entropy length in bits (default: 128)

        Raises:
            ValueError: If the bit length is not valid
        """
        if not self.IsValidEntropyBitLen(bit_len):
            raise ValueError(f"Entropy bit length is not valid ({bit_len})")
        super().__init__(bit_len)

    @staticmethod
    def IsValidEntropyBitLen(bit_len: int) -> bool:
        """
        Get if the specified entropy bit length is valid.

        Args:
            bit_len (int): Entropy length in bits

        Returns:
            bool: True if valid, false otherwise
        """
        return bit_len in ElectrumV1EntropyGeneratorConst.ENTROPY_BIT_LEN

    @staticmethod
    def IsValidEntropyByteLen(byte_len: int) -> bool:
        """
        Get if the specified entropy byte length is valid.

        Args:
            byte_len (int): Entropy length in bytes

        Returns:
            bool: True if valid, false otherwise
        """
        return ElectrumV1EntropyGenerator.IsValidEntropyBitLen(byte_len * 8)



# Copyright (c) 2022 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module containing utility classes for Electrum v1 keys derivation, since it uses its own algorithm."""

# Imports
from __future__ import annotations

from functools import lru_cache
from typing import Optional, Union

from bip_utils.addr import P2PKHAddr, P2PKHPubKeyModes
from bip_utils.bip.bip32 import Bip32KeyIndex
from bip_utils.coin_conf import CoinsConf
from bip_utils.ecc import IPrivateKey, IPublicKey, Secp256k1, Secp256k1PrivateKey, Secp256k1PublicKey
from bip_utils.utils.crypto import DoubleSha256
from bip_utils.utils.misc import AlgoUtils, BytesUtils, IntegerUtils


class ElectrumV1:
    """
    Electrum v1 class.
    It derives keys like the Electrum wallet with old (v1) mnemonic.
    """

    m_priv_key: Optional[IPrivateKey]
    m_pub_key: IPublicKey

    @classmethod
    def FromSeed(cls,
                 seed_bytes: bytes) -> ElectrumV1:
        """
        Construct class from seed bytes.

        Args:
            seed_bytes (bytes): Seed bytes

        Returns:
            ElectrumV1 object: ElectrumV1 object
        """

        # The seed is the private key itself
        return cls.FromPrivateKey(seed_bytes)

    @classmethod
    def FromPrivateKey(cls,
                       priv_key: Union[bytes, IPrivateKey]) -> ElectrumV1:
        """
        Construct class from private key.

        Args:
            priv_key (bytes or IPrivateKey): Private key

        Returns:
            ElectrumV1 object: ElectrumV1 object

        Raises:
            TypeError: if the private key is not a Secp256k1PrivateKey
        """
        if isinstance(priv_key, bytes):
            priv_key = Secp256k1PrivateKey.FromBytes(priv_key)
        return cls(priv_key, None)

    @classmethod
    def FromPublicKey(cls,
                      pub_key: Union[bytes, IPublicKey]) -> ElectrumV1:
        """
        Construct class from public key.

        Args:
            pub_key (bytes or IPublicKey): Public key

        Returns:
            ElectrumV1 object: ElectrumV1 object

        Raises:
            TypeError: if the public key is not a Secp256k1PublicKey
        """
        if isinstance(pub_key, bytes):
            pub_key = Secp256k1PublicKey.FromBytes(pub_key)
        return cls(None, pub_key)

    def __init__(self,
                 priv_key: Optional[IPrivateKey],
                 pub_key: Optional[IPublicKey]) -> None:
        """
        Construct class.

        Args:
            priv_key (IPrivateKey object, optional): Private key (None for a public-only object)
            pub_key (IPublicKey object, optional)  : Public key (only needed for a public-only object)
                                                     If priv_key is not None, it'll be discarded

        Raises:
            TypeError: if the private key is not a Secp256k1PrivateKey or the public key is not a Secp256k1PublicKey
        """
        if priv_key is not None:
            if not isinstance(priv_key, Secp256k1PrivateKey):
                raise TypeError("Private key shall be a secp256k1 key")

            self.m_priv_key = priv_key
            self.m_pub_key = priv_key.PublicKey()
        else:
            if not isinstance(pub_key, Secp256k1PublicKey):
                raise TypeError("Public key shall be a secp256k1 key")

            self.m_priv_key = None
            self.m_pub_key = pub_key

    def IsPublicOnly(self) -> bool:
        """
        Get if it's public-only.

        Returns:
            bool: True if public-only, false otherwise
        """
        return self.m_priv_key is None

    def MasterPrivateKey(self) -> IPrivateKey:
        """
        Get the master private key.

        Returns:
            IPrivateKey object: IPrivateKey object
        """
        if self.IsPublicOnly():
            raise ValueError("Public-only deterministic keys have no private half")

        assert isinstance(self.m_priv_key, Secp256k1PrivateKey)
        return self.m_priv_key

    def MasterPublicKey(self) -> IPublicKey:
        """
        Get the master public key.

        Returns:
            IPublicKey object: IPublicKey object
        """
        return self.m_pub_key

    def GetPrivateKey(self,
                      change_idx: int,
                      addr_idx: int) -> IPrivateKey:
        """
        Get the private key with the specified change and address indexes.
        Derivation path (not BIP32 derivation): m/change_idx/addr_idx

        Args:
            change_idx (int): Change index
            addr_idx (int)  : Address index

        Returns:
            IPrivateKey object: IPrivateKey object

        Raises:
            ValueError: If one of the index is not valid
        """
        if self.IsPublicOnly():
            raise ValueError("Public-only deterministic keys have no private half")
        return self.__DerivePrivateKey(change_idx, addr_idx)

    def GetPublicKey(self,
                     change_idx: int,
                     addr_idx: int) -> IPublicKey:
        """
        Get the public key with the specified change and address indexes.
        Derivation path (not BIP32 derivation): m/change_idx/addr_idx

        Args:
            change_idx (int): Change index
            addr_idx (int)  : Address index


        Returns:
            IPublicKey object: IPublicKey object

        Raises:
            ValueError: If one of the index is not valid
        """
        return (self.__DerivePublicKey(change_idx, addr_idx)
                if self.IsPublicOnly()
                else self.GetPrivateKey(change_idx, addr_idx).PublicKey())

    @lru_cache()
    def GetAddress(self,
                   change_idx: int,
                   addr_idx: int) -> str:
        """
        Get the address with the specified change and address indexes.
        Derivation path (not BIP32 derivation): m/change_idx/addr_idx

        Args:
            change_idx (int): Change index
            addr_idx (int)  : Address index

        Returns:
            str: Address

        Raises:
            ValueError: If one of the index is not valid
        """
        return P2PKHAddr.EncodeKey(self.GetPublicKey(change_idx, addr_idx),
                                   net_ver=CoinsConf.BitcoinMainNet.ParamByKey("p2pkh_net_ver"),
                                   pub_key_mode=P2PKHPubKeyModes.UNCOMPRESSED)

    @lru_cache()
    def __DerivePrivateKey(self,
                           change_idx: int,
                           addr_idx: int) -> IPrivateKey:
        """
        Derive the private key with the specified change and address indexes.
        Derivation path (not BIP32 derivation): m/change_idx/addr_idx

        Args:
            change_idx (int): Change index
            addr_idx (int)  : Address index

        Returns:
            IPrivateKey object: IPrivateKey object

        Raises:
            ValueError: If one of the index is not valid
        """
        self.__ValidateIndexes(change_idx, addr_idx)

        seq_bytes = self.__GetSequence(change_idx, addr_idx)
        priv_key_int = (self.MasterPrivateKey().Raw().ToInt() + BytesUtils.ToInteger(seq_bytes)) % Secp256k1.Order()
        return Secp256k1PrivateKey.FromBytes(
            IntegerUtils.ToBytes(priv_key_int, Secp256k1PrivateKey.Length())
        )

    @lru_cache()
    def __DerivePublicKey(self,
                          change_idx: int,
                          addr_idx: int) -> IPublicKey:
        """
        Derive the public key with the specified change and address indexes.
        Derivation path (not BIP32 derivation): m/change_idx/addr_idx

        Args:
            change_idx (int): Change index
            addr_idx (int)  : Address index

        Returns:
            IPublicKey object: IPublicKey object

        Raises:
            ValueError: If one of the index is not valid
        """
        self.__ValidateIndexes(change_idx, addr_idx)

        seq_bytes = self.__GetSequence(change_idx, addr_idx)
        return Secp256k1PublicKey.FromPoint(
            self.MasterPublicKey().Point() + BytesUtils.ToInteger(seq_bytes) * Secp256k1.Generator()
        )

    def __GetSequence(self,
                      change_idx: int,
                      addr_idx: int) -> bytes:
        """
        Get sequence.

        Args:
            change_idx (int): Change index
            addr_idx (int)  : Address index

        Returns:
            bytes: Sequence bytes
        """
        return DoubleSha256.QuickDigest(
            AlgoUtils.Encode(f"{addr_idx}:{change_idx}:") + self.MasterPublicKey().RawUncompressed().ToBytes()[1:])

    @staticmethod
    def __ValidateIndexes(change_idx: int,
                          addr_idx: int) -> None:
        """
        Validate indexes and raise a ValueError if not valid.

        Args:
            change_idx (int): Change index
            addr_idx (int)  : Address index

        Raises:
            ValueError: If one of the index is not valid
        """

        # Just try to create a key index object
        Bip32KeyIndex(change_idx)
        Bip32KeyIndex(addr_idx)



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Module for ed25519-kholaw keys.
With respect to ed25519, the private key has a length of 64-byte (left 32-byte of the ed25519 private key and a
right 32-byte extension part).
"""

# Imports
from typing import Any

from nacl import signing

from bip_utils.ecc.common.ikeys import IPrivateKey, IPublicKey
from bip_utils.ecc.common.ipoint import IPoint
from bip_utils.ecc.curve.elliptic_curve_types import EllipticCurveTypes
from bip_utils.ecc.ed25519.ed25519_keys import Ed25519PrivateKey, Ed25519PublicKey
from bip_utils.ecc.ed25519.lib import ed25519_lib
from bip_utils.ecc.ed25519_kholaw.ed25519_kholaw_point import Ed25519KholawPoint
from bip_utils.utils.misc import DataBytes


class Ed25519KholawKeysConst:
    """Class container for ed25519-kholaw keys constants."""

    # Private key length in bytes
    PRIV_KEY_BYTE_LEN: int = 64


class Ed25519KholawPublicKey(Ed25519PublicKey):
    """Ed25519-Kholaw public key class."""

    @staticmethod
    def CurveType() -> EllipticCurveTypes:
        """
        Get the elliptic curve type.

        Returns:
           EllipticCurveTypes: Elliptic curve type
        """
        return EllipticCurveTypes.ED25519_KHOLAW

    def Point(self) -> IPoint:
        """
        Get public key point.

        Returns:
            IPoint object: IPoint object
        """
        return Ed25519KholawPoint(bytes(self.m_ver_key))


class Ed25519KholawPrivateKey(IPrivateKey):
    """Ed25519-Kholaw private key class."""

    m_sign_key: Ed25519PrivateKey
    m_ext_key: bytes

    @classmethod
    def FromBytes(cls,
                  key_bytes: bytes) -> IPrivateKey:
        """
        Construct class from key bytes.

        Args:
            key_bytes (bytes): Key bytes

        Returns:
            IPrivateKey: IPrivateKey object

        Raises:
            ValueError: If key bytes are not valid
        """
        return cls(Ed25519PrivateKey.FromBytes(key_bytes[:Ed25519PrivateKey.Length()]),
                   key_bytes[Ed25519PrivateKey.Length():])

    def __init__(self,
                 key_obj: IPrivateKey,
                 key_ex_bytes: bytes) -> None:
        """
        Construct class.

        Args:
            key_obj (IPrivateKey object): Key object, shall be an Ed25519PrivateKey private key
            key_ex_bytes (bytes)        : Key extended bytes

        Raises:
            TypeError: If key object is not of the correct type
            ValueError: If extended key is not valid
        """
        if not isinstance(key_obj, Ed25519PrivateKey):
            raise TypeError("Invalid private key object type")
        if len(key_ex_bytes) != Ed25519PrivateKey.Length():
            raise ValueError("Invalid extended key length")

        self.m_sign_key = key_obj
        self.m_ext_key = key_ex_bytes

    @staticmethod
    def CurveType() -> EllipticCurveTypes:
        """
        Get the elliptic curve type.

        Returns:
           EllipticCurveTypes: Elliptic curve type
        """
        return EllipticCurveTypes.ED25519_KHOLAW

    @staticmethod
    def Length() -> int:
        """
        Get the key length.

        Returns:
           int: Key length
        """
        return Ed25519KholawKeysConst.PRIV_KEY_BYTE_LEN

    def UnderlyingObject(self) -> Any:
        """
        Get the underlying object.

        Returns:
           Any: Underlying object
        """
        return self.m_sign_key.UnderlyingObject()

    def Raw(self) -> DataBytes:
        """
        Return raw private key.

        Returns:
            DataBytes object: DataBytes object
        """
        return DataBytes(self.m_sign_key.Raw().ToBytes() + self.m_ext_key)

    def PublicKey(self) -> IPublicKey:
        """
        Get the public key correspondent to the private one.

        Returns:
            IPublicKey object: IPublicKey object
        """
        return Ed25519KholawPublicKey(
            signing.VerifyKey(
                ed25519_lib.point_scalar_mul_base(bytes(self.m_sign_key.UnderlyingObject()))
            )
        )



# Copyright (c) 2022 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module for Electrum v2 mnemonic generation."""

# Imports
from typing import Optional

from bip_utils.bip.bip39 import Bip39MnemonicValidator
from bip_utils.electrum.mnemonic_v1 import ElectrumV1MnemonicValidator
from bip_utils.electrum.mnemonic_v2.electrum_v2_mnemonic import ElectrumV2MnemonicConst, ElectrumV2MnemonicTypes
from bip_utils.utils.crypto import HmacSha512
from bip_utils.utils.misc import BytesUtils
from bip_utils.utils.mnemonic import Mnemonic


class ElectrumV2MnemonicUtilsConst:
    """Class container for Electrum v2 mnemonic utility constants."""

    # HMAC key
    HMAC_KEY: bytes = b"Seed version"


class ElectrumV2MnemonicUtils:
    """Class container for Electrum v2 mnemonic utility functions."""

    @staticmethod
    def IsValidMnemonic(mnemonic: Mnemonic,
                        mnemonic_type: Optional[ElectrumV2MnemonicTypes] = None) -> bool:
        """
        Get if the specified mnemonic is valid.

        Args:
            mnemonic (Mnemonic)                    : Mnemonic
            mnemonic_type (ElectrumV2MnemonicTypes): Mnemonic type

        Returns:
            bool: True if valid, false otherwise
        """
        if ElectrumV2MnemonicUtils.__IsBip39OrV1Mnemonic(mnemonic):
            return False
        return (ElectrumV2MnemonicUtils.__IsType(mnemonic, mnemonic_type)
                if mnemonic_type is not None
                else ElectrumV2MnemonicUtils.__IsAnyType(mnemonic))

    @staticmethod
    def __IsBip39OrV1Mnemonic(mnemonic: Mnemonic) -> bool:
        """
        Get if the specified mnemonic is a valid BIP39 or v1 Electrum mnemonic.

        Args:
            mnemonic (Mnemonic): Mnemonic

        Returns:
            bool: True if valid, false otherwise
        """
        return Bip39MnemonicValidator().IsValid(mnemonic) or ElectrumV1MnemonicValidator().IsValid(mnemonic)

    @staticmethod
    def __IsAnyType(mnemonic: Mnemonic) -> bool:
        """
        Get if the specified mnemonic is of any valid type.

        Args:
            mnemonic (Mnemonic): Mnemonic

        Returns:
            bool: True if valid, false otherwise
        """
        h = HmacSha512.QuickDigest(ElectrumV2MnemonicUtilsConst.HMAC_KEY, mnemonic.ToStr())
        for mnemonic_type in ElectrumV2MnemonicTypes:
            if BytesUtils.ToHexString(h).startswith(ElectrumV2MnemonicConst.TYPE_TO_PREFIX[mnemonic_type]):
                return True
        return False

    @staticmethod
    def __IsType(mnemonic: Mnemonic,
                 mnemonic_type: ElectrumV2MnemonicTypes) -> bool:
        """
        Get if the specified mnemonic is of the specified type.

        Args:
            mnemonic (Mnemonic)                    : Mnemonic
            mnemonic_type (ElectrumV2MnemonicTypes): Mnemonic type

        Returns:
            bool: True if valid, false otherwise
        """
        h = HmacSha512.QuickDigest(ElectrumV2MnemonicUtilsConst.HMAC_KEY, mnemonic.ToStr())
        return BytesUtils.ToHexString(h).startswith(ElectrumV2MnemonicConst.TYPE_TO_PREFIX[mnemonic_type])



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module for Monero mnemonic."""

# Imports
from enum import IntEnum, auto, unique
from typing import Dict, List

from bip_utils.utils.mnemonic import Mnemonic, MnemonicLanguages


@unique
class MoneroWordsNum(IntEnum):
    """Enumerative for Monero words number."""

    WORDS_NUM_12 = 12   # No checksum
    WORDS_NUM_13 = 13   # With checksum
    WORDS_NUM_24 = 24   # No checksum
    WORDS_NUM_25 = 25   # With checksum


@unique
class MoneroLanguages(MnemonicLanguages):
    """Enumerative for Monero languages."""

    CHINESE_SIMPLIFIED = auto()
    DUTCH = auto()
    ENGLISH = auto()
    FRENCH = auto()
    GERMAN = auto()
    ITALIAN = auto()
    JAPANESE = auto()
    PORTUGUESE = auto()
    SPANISH = auto()
    RUSSIAN = auto()


class MoneroMnemonicConst:
    """Class container for Monero mnemonic constants."""

    # Accepted mnemonic word numbers
    MNEMONIC_WORD_NUM: List[MoneroWordsNum] = [
        MoneroWordsNum.WORDS_NUM_12,
        MoneroWordsNum.WORDS_NUM_13,
        MoneroWordsNum.WORDS_NUM_24,
        MoneroWordsNum.WORDS_NUM_25,
    ]

    # Mnemonic word numbers with checksum
    MNEMONIC_WORD_NUM_CHKSUM: List[MoneroWordsNum] = [
        MoneroWordsNum.WORDS_NUM_13,
        MoneroWordsNum.WORDS_NUM_25,
    ]

    # Language unique prefix lengths
    LANGUAGE_UNIQUE_PREFIX_LEN: Dict[MnemonicLanguages, int] = {
        MoneroLanguages.CHINESE_SIMPLIFIED: 1,
        MoneroLanguages.DUTCH: 4,
        MoneroLanguages.ENGLISH: 3,
        MoneroLanguages.FRENCH: 4,
        MoneroLanguages.GERMAN: 4,
        MoneroLanguages.ITALIAN: 4,
        MoneroLanguages.JAPANESE: 4,
        MoneroLanguages.PORTUGUESE: 4,
        MoneroLanguages.SPANISH: 4,
        MoneroLanguages.RUSSIAN: 4,
    }

    # Language files
    LANGUAGE_FILES: Dict[MnemonicLanguages, str] = {
        MoneroLanguages.CHINESE_SIMPLIFIED: "wordlist/chinese_simplified.txt",
        MoneroLanguages.DUTCH: "wordlist/dutch.txt",
        MoneroLanguages.ENGLISH: "wordlist/english.txt",
        MoneroLanguages.FRENCH: "wordlist/french.txt",
        MoneroLanguages.GERMAN: "wordlist/german.txt",
        MoneroLanguages.ITALIAN: "wordlist/italian.txt",
        MoneroLanguages.JAPANESE: "wordlist/japanese.txt",
        MoneroLanguages.PORTUGUESE: "wordlist/portuguese.txt",
        MoneroLanguages.SPANISH: "wordlist/spanish.txt",
        MoneroLanguages.RUSSIAN: "wordlist/russian.txt",
    }

    # Total number of words
    WORDS_LIST_NUM: int = 1626


class MoneroMnemonic(Mnemonic):
    """Monero mnemonic class (alias for Mnemonic)."""



from bip_utils.algorand.mnemonic.algorand_entropy_generator import AlgorandEntropyBitLen, AlgorandEntropyGenerator
from bip_utils.algorand.mnemonic.algorand_mnemonic import AlgorandLanguages, AlgorandMnemonic, AlgorandWordsNum
from bip_utils.algorand.mnemonic.algorand_mnemonic_decoder import AlgorandMnemonicDecoder
from bip_utils.algorand.mnemonic.algorand_mnemonic_encoder import AlgorandMnemonicEncoder
from bip_utils.algorand.mnemonic.algorand_mnemonic_generator import AlgorandMnemonicGenerator
from bip_utils.algorand.mnemonic.algorand_mnemonic_validator import AlgorandMnemonicValidator
from bip_utils.algorand.mnemonic.algorand_seed_generator import AlgorandSeedGenerator



# Copyright (c) 2022 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Module for Cardano Icarus master key generation.

References:
    https://input-output-hk.github.io/cardano-wallet/concepts/master-key-generation
    https://cips.cardano.org/cips/cip3/icarus.md
"""

# Imports
from typing import Tuple

from bip_utils.bip.bip32 import IBip32MstKeyGenerator
from bip_utils.bip.bip32.slip10.bip32_slip10_mst_key_generator import Bip32Slip10MstKeyGeneratorConst
from bip_utils.ecc import Ed25519KholawPrivateKey
from bip_utils.utils.crypto import Pbkdf2HmacSha512
from bip_utils.utils.misc import BitUtils


class CardanoIcarusMasterKeyGeneratorConst:
    """Class container for Cardano Icarus master key generator constants."""

    # PBKDF2 password
    PBKDF2_PASSWORD: str = ""
    # PBKDF2 rounds
    PBKDF2_ROUNDS: int = 4096
    # PBKDF2 output length in bytes
    PBKDF2_OUT_BYTE_LEN: int = 96


class CardanoIcarusMstKeyGenerator(IBip32MstKeyGenerator):
    """
    Cardano Icarus master key generator class.
    It allows master keys generation in according to Cardano Icarus.
    """

    @classmethod
    def GenerateFromSeed(cls,
                         seed_bytes: bytes) -> Tuple[bytes, bytes]:
        """
        Generate a master key from the specified seed.

        Args:
            seed_bytes (bytes): Seed bytes

        Returns:
            tuple[bytes, bytes]: Private key bytes (index 0) and chain code bytes (index 1)

        Raises:
            Bip32KeyError: If the seed is not suitable for master key generation
            ValueError: If seed length is not valid
        """
        if len(seed_bytes) < Bip32Slip10MstKeyGeneratorConst.SEED_MIN_BYTE_LEN:
            raise ValueError(f"Invalid seed length ({len(seed_bytes)})")

        key_bytes = Pbkdf2HmacSha512.DeriveKey(CardanoIcarusMasterKeyGeneratorConst.PBKDF2_PASSWORD,
                                               seed_bytes,
                                               CardanoIcarusMasterKeyGeneratorConst.PBKDF2_ROUNDS,
                                               CardanoIcarusMasterKeyGeneratorConst.PBKDF2_OUT_BYTE_LEN)
        key_bytes = cls.__TweakMasterKeyBits(key_bytes)

        return key_bytes[:Ed25519KholawPrivateKey.Length()], key_bytes[Ed25519KholawPrivateKey.Length():]

    @staticmethod
    def __TweakMasterKeyBits(key_bytes: bytes) -> bytes:
        """
        Tweak master key bits.

        Args:
            key_bytes (bytes): Key bytes

        Returns:
            bytes: Tweaked key bytes
        """
        key_bytes = bytearray(key_bytes)
        # Clear the lowest 3 bits of the first byte of kL
        key_bytes[0] = BitUtils.ResetBits(key_bytes[0], 0x07)
        # Clear the highest 3 bits of the last byte of kL (standard kholaw only clears the highest one)
        key_bytes[31] = BitUtils.ResetBits(key_bytes[31], 0xE0)
        # Set the second-highest bit of the last byte of kL
        key_bytes[31] = BitUtils.SetBits(key_bytes[31], 0x40)

        return bytes(key_bytes)



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module with helper class for Substrate coins configuration handling."""

# Imports
from __future__ import annotations

from typing import Dict

from bip_utils.coin_conf import CoinConf
from bip_utils.utils.conf import CoinNames as UtilsCoinNames


class SubstrateCoinConf:
    """Substrate coin configuration class."""

    m_coin_names: UtilsCoinNames
    m_ss58_format: int
    m_addr_params: Dict[str, int]

    @classmethod
    def FromCoinConf(cls,
                     coin_conf: CoinConf) -> SubstrateCoinConf:
        """
        Construct class.

        Args:
            coin_conf (CoinConf object): Generic coin configuration object

        Returns:
            SubstrateCoinConf object: SubstrateCoinConf object
        """
        return cls(coin_names=coin_conf.CoinNames(),
                   ss58_format=coin_conf.ParamByKey("addr_ss58_format"))

    def __init__(self,
                 coin_names: UtilsCoinNames,
                 ss58_format: int) -> None:
        """
        Construct class.

        Args:
            coin_names (CoinNames object): Coin names
            ss58_format (int)            : SS58 format
        """
        self.m_coin_names = coin_names
        self.m_ss58_format = ss58_format
        self.m_addr_params = {"ss58_format": ss58_format}

    def CoinNames(self) -> UtilsCoinNames:
        """
        Get coin names.

        Returns:
            CoinNames object: CoinNames object
        """
        return self.m_coin_names

    def SS58Format(self) -> int:
        """
        Get SS58 format.

        Returns:
            int: SS58 format
        """
        return self.m_ss58_format

    def AddrParams(self) -> Dict[str, int]:
        """
        Get the address parameters.

        Returns:
            dict: Address parameters
        """
        return self.m_addr_params



# Copyright (c) 2022 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Module for CBOR decoding/encoding indefinite length arrays.
Indefinite length arrays are encoded without writing the array length, so elements shall be read until
the termination byte is found.

NOTE: encoding of values greater than 2^64 is not supported.
"""

# Imports
from enum import IntEnum, unique
from typing import Dict, List, Sequence

import cbor2

from bip_utils.utils.misc.integer import IntegerUtils


@unique
class CborIds(IntEnum):
    """Enumerative for CBOR identifiers."""

    UINT8 = 24
    UINT16 = 25
    UINT32 = 26
    UINT64 = 27
    INDEF_LEN_ARRAY_START = 0x9F
    INDEF_LEN_ARRAY_END = 0xFF


class CborIndefiniteLenArrayConst:
    """Class container for CBOR indefinite length arrays constants."""

    # CBOR uint IDs to byte length
    UINT_IDS_TO_BYTE_LEN: Dict[int, int] = {
        CborIds.UINT8: 2,
        CborIds.UINT16: 3,
        CborIds.UINT32: 5,
        CborIds.UINT64: 9,
    }


class CborIndefiniteLenArrayDecoder:
    """
    CBOR indefinite length arrays decoder.
    It decodes bytes back to array.
    """

    @staticmethod
    def Decode(enc_bytes: bytes) -> List[int]:
        """
        CBOR-decode the specified bytes.

        Args:
            enc_bytes (bytes): Encoded bytes

        Returns:
            list[int]: List of integers

        Raises:
            ValueError: If encoding is not valid
        """

        # Check for validity
        if len(enc_bytes) < 3:
            raise ValueError(f"Invalid length ({len(enc_bytes)})")
        if enc_bytes[0] != CborIds.INDEF_LEN_ARRAY_START:
            raise ValueError(f"Invalid first byte ({enc_bytes[0]})")
        if enc_bytes[-1] != CborIds.INDEF_LEN_ARRAY_END:
            raise ValueError(f"Invalid last byte ({enc_bytes[-1]})")

        # Continue to decode elements until the end value is found
        i = 1
        int_elems = []
        while True:
            # Get current byte
            if i >= len(enc_bytes):
                raise ValueError("Invalid encoding (index overflow)")
            curr_val = enc_bytes[i]
            if curr_val == CborIds.INDEF_LEN_ARRAY_END:
                break
            # Get current length (1-byte if ID is not found)
            curr_len = CborIndefiniteLenArrayConst.UINT_IDS_TO_BYTE_LEN.get(curr_val, 1)
            # CBOR-decode the current integer
            int_elems.append(cbor2.loads(enc_bytes[i:i + curr_len]))
            # Move forward
            i += curr_len

        return int_elems


class CborIndefiniteLenArrayEncoder:
    """
    CBOR indefinite length arrays encoder.
    It encodes indefinite length arrays to bytes.
    """

    @staticmethod
    def Encode(int_seq: Sequence[int]) -> bytes:
        """
        CBOR-encode the specified elements.

        Args:
            int_seq (sequence[int]): Collection of integers

        Returns:
            bytes: CBOR-encoded bytes
        """
        return (IntegerUtils.ToBytes(CborIds.INDEF_LEN_ARRAY_START, bytes_num=1)
                + b"".join([cbor2.dumps(p) for p in int_seq])
                + IntegerUtils.ToBytes(CborIds.INDEF_LEN_ARRAY_END, bytes_num=1))



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module for generic entropy generator."""

# Imports
import os
import secrets

from bip_utils.utils.misc import IntegerUtils


class EntropyGenerator:
    """
    Entropy generator class.
    It generates random entropy bytes with the specified length.
    """

    m_bit_len: int

    def __init__(self,
                 bit_len: int) -> None:
        """
        Construct class.

        Args:
            bit_len (int): Entropy length in bits
        """
        self.m_bit_len = bit_len

    def Generate(self) -> bytes:
        """
        Generate random entropy bytes.

        Returns:
            bytes: Generated entropy bytes
        """
        return (os.urandom(self.m_bit_len // 8)
                if self.m_bit_len % 8 == 0
                else IntegerUtils.ToBytes(secrets.randbits(self.m_bit_len)))



# Copyright (c) 2022 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module for Electrum v1 mnemonic."""

# Imports
from enum import IntEnum, auto, unique
from typing import Dict, List

from bip_utils.bip.bip39 import Bip39Mnemonic
from bip_utils.utils.mnemonic import MnemonicLanguages


@unique
class ElectrumV1WordsNum(IntEnum):
    """Enumerative for Electrum words number (v1)."""

    WORDS_NUM_12 = 12


@unique
class ElectrumV1Languages(MnemonicLanguages):
    """Enumerative for Electrum languages (v1)."""

    ENGLISH = auto()


class ElectrumV1MnemonicConst:
    """Class container for Electrum v1 mnemonic constants."""

    # Accepted mnemonic word numbers
    MNEMONIC_WORD_NUM: List[ElectrumV1WordsNum] = [
        ElectrumV1WordsNum.WORDS_NUM_12,
    ]

    # Language files
    LANGUAGE_FILES: Dict[MnemonicLanguages, str] = {
        ElectrumV1Languages.ENGLISH: "wordlist/english.txt",
    }

    # Total number of words
    WORDS_LIST_NUM: int = 1626


class ElectrumV1Mnemonic(Bip39Mnemonic):
    """Electrum v1 mnemonic class."""



from bip_utils.cardano.mnemonic.cardano_byron_legacy_seed_generator import CardanoByronLegacySeedGenerator
from bip_utils.cardano.mnemonic.cardano_icarus_seed_generator import CardanoIcarusSeedGenerator



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module with helper class for Monero coins configuration handling."""

# Imports
from __future__ import annotations

from typing import Dict

from bip_utils.coin_conf import CoinConf
from bip_utils.utils.conf import CoinNames as UtilsCoinNames


class MoneroCoinConf:
    """Monero coin configuration class."""

    m_coin_names: UtilsCoinNames
    m_addr_net_ver: bytes
    m_int_addr_net_ver: bytes
    m_subaddr_net_ver: bytes
    m_addr_params: Dict[str, bytes]

    @classmethod
    def FromCoinConf(cls,
                     coin_conf: CoinConf) -> MoneroCoinConf:
        """
        Construct class.

        Args:
            coin_conf (CoinConf object): Generic coin configuration object

        Returns:
            MoneroCoinConf object: MoneroCoinConf object
        """
        return cls(coin_names=coin_conf.CoinNames(),
                   addr_net_ver=coin_conf.ParamByKey("addr_net_ver"),
                   int_addr_net_ver=coin_conf.ParamByKey("addr_int_net_ver"),
                   subaddr_net_ver=coin_conf.ParamByKey("subaddr_net_ver"))

    def __init__(self,
                 coin_names: UtilsCoinNames,
                 addr_net_ver: bytes,
                 int_addr_net_ver: bytes,
                 subaddr_net_ver: bytes) -> None:
        """
        Construct class.

        Args:
            coin_names (CoinNames object): Coin names
            addr_net_ver (bytes)         : Address net version
            int_addr_net_ver (bytes)     : Integrated address net version
            subaddr_net_ver (bytes)      : Subaddress net version
        """
        self.m_coin_names = coin_names
        self.m_addr_net_ver = addr_net_ver
        self.m_int_addr_net_ver = int_addr_net_ver
        self.m_subaddr_net_ver = subaddr_net_ver

    def CoinNames(self) -> UtilsCoinNames:
        """
        Get coin names.

        Returns:
            CoinNames object: CoinNames object
        """
        return self.m_coin_names

    def AddrNetVersion(self) -> bytes:
        """
        Get address net version.

        Returns:
            bytes: Address net version
        """
        return self.m_addr_net_ver

    def IntegratedAddrNetVersion(self) -> bytes:
        """
        Get integrated address net version.

        Returns:
            bytes: Address net version
        """
        return self.m_int_addr_net_ver

    def SubaddrNetVersion(self) -> bytes:
        """
        Get subaddress net version.

        Returns:
            bytes: Subaddress net version
        """
        return self.m_subaddr_net_ver



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module with interface for BIP39 seed generation classes."""

# Imports
from abc import ABC, abstractmethod
from typing import Optional, Union

from bip_utils.bip.bip39.bip39_mnemonic import Bip39Languages
from bip_utils.utils.mnemonic import Mnemonic


class IBip39SeedGenerator(ABC):
    """BIP39 seed generator interface."""

    @abstractmethod
    def __init__(self,
                 mnemonic: Union[str, Mnemonic],
                 lang: Optional[Bip39Languages]) -> None:
        """
        Construct class.

        Args:
            mnemonic (str or Mnemonic object): Mnemonic
            lang (Bip39Languages, optional)  : Language, None for automatic detection

        Raises:
            ValueError: If the mnemonic is not valid
        """

    @abstractmethod
    def Generate(self,
                 passphrase: str) -> bytes:
        """
        Generate the seed using the specified passphrase.

        Args:
            passphrase (str, optional): Passphrase, empty if not specified

        Returns:
            bytes: Generated seed
        """



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module for BIP39 mnemonic generation."""

# Imports
from typing import Union

from bip_utils.bip.bip39.bip39_entropy_generator import Bip39EntropyGenerator
from bip_utils.bip.bip39.bip39_mnemonic import Bip39Languages, Bip39MnemonicConst, Bip39WordsNum
from bip_utils.bip.bip39.bip39_mnemonic_encoder import Bip39MnemonicEncoder
from bip_utils.utils.mnemonic import Mnemonic


class Bip39MnemonicGenerator:
    """
    BIP39 mnemonic generator class. It generates mnemonics in according to BIP39.
    Mnemonic can be generated randomly from words number or from a specified entropy.
    """

    m_mnemonic_encoder: Bip39MnemonicEncoder

    def __init__(self,
                 lang: Bip39Languages = Bip39Languages.ENGLISH) -> None:
        """
        Construct class.

        Args:
            lang (Bip39Languages, optional): Language (default: English)

        Raises:
            TypeError: If the language is not a Bip39Languages enum
            ValueError: If language words list is not valid
        """
        self.m_mnemonic_encoder = Bip39MnemonicEncoder(lang)

    def FromWordsNumber(self,
                        words_num: Union[int, Bip39WordsNum]) -> Mnemonic:
        """
        Generate mnemonic with the specified words number from random entropy.

        Args:
            words_num (int or Bip39WordsNum): Number of words (12, 15, 18, 21, 24)

        Returns:
            Mnemonic object: Generated mnemonic

        Raises:
            ValueError: If words number is not valid
        """

        # Check words number
        if words_num not in Bip39MnemonicConst.MNEMONIC_WORD_NUM:
            raise ValueError(f"Words number for mnemonic ({words_num}) is not valid")

        # Get entropy length in bit from words number
        entropy_bit_len = self.__EntropyBitLenFromWordsNum(words_num)
        # Generate entropy
        entropy_bytes = Bip39EntropyGenerator(entropy_bit_len).Generate()

        return self.FromEntropy(entropy_bytes)

    def FromEntropy(self,
                    entropy_bytes: bytes) -> Mnemonic:
        """
        Generate mnemonic from the specified entropy bytes.

        Args:
            entropy_bytes (bytes): Entropy bytes (accepted lengths in bits: 128, 160, 192, 224, 256)

        Returns:
            Mnemonic object: Generated mnemonic

        Raises:
            ValueError: If entropy byte length is not valid
        """
        return self.m_mnemonic_encoder.Encode(entropy_bytes)

    @staticmethod
    def __EntropyBitLenFromWordsNum(words_num: int) -> int:
        """
        Get entropy length from words number.

        Args:
            words_num (int): Words number

        Returns:
            int: Correspondent entropy length
        """
        return (words_num * Bip39MnemonicConst.WORD_BIT_LEN) - (words_num // 3)



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module for Ripple address encoding/decoding."""

# Imports
from typing import Any, Union

from bip_utils.addr.iaddr_decoder import IAddrDecoder
from bip_utils.addr.iaddr_encoder import IAddrEncoder
from bip_utils.addr.P2PKH_addr import P2PKHAddrDecoder, P2PKHAddrEncoder
from bip_utils.base58 import Base58Alphabets
from bip_utils.coin_conf import CoinsConf
from bip_utils.ecc import IPublicKey


class XrpAddrDecoder(IAddrDecoder):
    """
    Ripple address decoder class.
    It allows the Ripple address decoding.
    """

    @staticmethod
    def DecodeAddr(addr: str,
                   **kwargs: Any) -> bytes:
        """
        Decode a Ripple address to bytes.

        Args:
            addr (str): Address string
            **kwargs  : Not used

        Returns:
            bytes: Public key hash bytes

        Raises:
            ValueError: If the address encoding is not valid
        """

        # Ripple address is just a P2PKH address with a different Base58 alphabet
        return P2PKHAddrDecoder.DecodeAddr(addr,
                                           net_ver=CoinsConf.Ripple.ParamByKey("p2pkh_net_ver"),
                                           base58_alph=Base58Alphabets.RIPPLE)


class XrpAddrEncoder(IAddrEncoder):
    """
    Ripple address encoder class.
    It allows the Ripple address encoding.
    """

    @staticmethod
    def EncodeKey(pub_key: Union[bytes, IPublicKey],
                  **kwargs: Any) -> str:
        """
        Encode a public key to Ripple address.

        Args:
            pub_key (bytes or IPublicKey): Public key bytes or object
            **kwargs                     : Not used

        Returns:
            str: Address string

        Raises:
            ValueError: If the public key is not valid
            TypeError: If the public key is not secp256k1
        """

        # Ripple address is just a P2PKH address with a different Base58 alphabet
        return P2PKHAddrEncoder.EncodeKey(pub_key,
                                          net_ver=CoinsConf.Ripple.ParamByKey("p2pkh_net_ver"),
                                          base58_alph=Base58Alphabets.RIPPLE)


# Deprecated: only for compatibility, Encoder class shall be used instead
XrpAddr = XrpAddrEncoder



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Module for BIP38 encryption/decryption.
Reference: https://github.com/bitcoin/bips/blob/master/bip-0038.mediawiki
"""

# Imports
from typing import Tuple, Union

from bip_utils.base58 import Base58Decoder, Base58Encoder
from bip_utils.bip.bip38.bip38_addr import Bip38Addr, Bip38PubKeyModes
from bip_utils.ecc import IPrivateKey, Secp256k1PrivateKey
from bip_utils.utils.crypto import AesEcbDecrypter, AesEcbEncrypter, Scrypt
from bip_utils.utils.misc import BytesUtils, IntegerUtils, StringUtils


class Bip38NoEcConst:
    """Class container for BIP38 no EC constants."""

    # Encrypted key byte length
    ENC_KEY_BYTE_LEN: int = 39
    # Encrypted key prefix
    ENC_KEY_PREFIX: bytes = b"\x01\x42"
    # Flagbyte for compressed public key
    FLAGBYTE_COMPRESSED: bytes = b"\xe0"
    # Flagbyte for uncompressed public key
    FLAGBYTE_UNCOMPRESSED: bytes = b"\xc0"
    # Parameters for scrypt algorithm
    SCRYPT_KEY_LEN: int = 64
    SCRYPT_N: int = 16384
    SCRYPT_P: int = 8
    SCRYPT_R: int = 8


class _Bip38NoEcUtils:
    """Class container for BIP38 no EC utility functions."""

    @staticmethod
    def AddressHash(priv_key_bytes: bytes,
                    pub_key_mode: Bip38PubKeyModes) -> bytes:
        """
        Compute the address hash as specified in BIP38 (without EC multiplication).

        Args:
            priv_key_bytes (bytes)         : private key bytes
            pub_key_mode (Bip38PubKeyModes): Public key mode

        Returns:
            bytes: Address hash

        Raises:
            ValueError: If the private key is not valid
        """
        return Bip38Addr.AddressHash(Secp256k1PrivateKey.FromBytes(priv_key_bytes).PublicKey(),
                                     pub_key_mode)

    @staticmethod
    def DeriveKeyHalves(passphrase: str,
                        address_hash: bytes) -> Tuple[bytes, bytes]:
        """
        Compute the scrypt as specified in BIP38 (without EC multiplication) and derive the two key halves.

        Args:
            passphrase (str)    : Passphrase
            address_hash (bytes): Address hash

        Returns:
            tuple[bytes, bytes]: Derived key halves
        """

        # Derive a key from passphrase and address hash
        key = Scrypt.DeriveKey(StringUtils.NormalizeNfc(passphrase),
                               address_hash,
                               key_len=Bip38NoEcConst.SCRYPT_KEY_LEN,
                               n=Bip38NoEcConst.SCRYPT_N,
                               r=Bip38NoEcConst.SCRYPT_R,
                               p=Bip38NoEcConst.SCRYPT_P)
        # Split the resulting 64 bytes in half
        derived_half_1 = key[:Bip38NoEcConst.SCRYPT_KEY_LEN // 2]
        derived_half_2 = key[Bip38NoEcConst.SCRYPT_KEY_LEN // 2:]

        return derived_half_1, derived_half_2


class Bip38NoEcEncrypter:
    """
    BIP38 encrypter class.
    It encrypts a private key using the algorithm specified in BIP38 without EC multiplication.
    """

    @staticmethod
    def Encrypt(priv_key: Union[bytes, IPrivateKey],
                passphrase: str,
                pub_key_mode: Bip38PubKeyModes) -> str:
        """
        Encrypt the specified private key.

        Args:
            priv_key (bytes or IPrivateKey): Private key bytes or object
            passphrase (str)               : Passphrase
            pub_key_mode (Bip38PubKeyModes): Public key mode

        Returns:
            str: Encrypted private key

        Raises:
            TypeError: If the private key is not a Secp256k1PrivateKey
            ValueError: If the private key bytes are not valid
        """

        # Convert to private key to check if bytes are valid
        if isinstance(priv_key, bytes):
            priv_key = Secp256k1PrivateKey.FromBytes(priv_key)
        elif not isinstance(priv_key, Secp256k1PrivateKey):
            raise TypeError("A secp256k1 private key is required")

        # Compute address hash
        priv_key_bytes = priv_key.Raw().ToBytes()
        address_hash = _Bip38NoEcUtils.AddressHash(priv_key_bytes, pub_key_mode)

        # Derive key halves from the passphrase and address hash
        derived_half_1, derived_half_2 = _Bip38NoEcUtils.DeriveKeyHalves(passphrase, address_hash)
        # Encrypt private key in two halves
        encrypted_half_1, encrypted_half_2 = Bip38NoEcEncrypter.__EncryptPrivateKey(priv_key_bytes,
                                                                                    derived_half_1,
                                                                                    derived_half_2)

        # Get flagbyte
        flagbyte = (Bip38NoEcConst.FLAGBYTE_COMPRESSED
                    if pub_key_mode == Bip38PubKeyModes.COMPRESSED
                    else Bip38NoEcConst.FLAGBYTE_UNCOMPRESSED)

        # Concatenate all parts
        enc_key_bytes = Bip38NoEcConst.ENC_KEY_PREFIX + flagbyte + address_hash + encrypted_half_1 + encrypted_half_2

        # Encode in Base58Check
        return Base58Encoder.CheckEncode(enc_key_bytes)

    @staticmethod
    def __EncryptPrivateKey(priv_key_bytes: bytes,
                            derived_half_1: bytes,
                            derived_half_2: bytes) -> Tuple[bytes, bytes]:
        """
        Encrypt private key in two halves.

        Args:
            priv_key_bytes (bytes): Private key
            derived_half_1 (bytes): First half of derived key
            derived_half_2 (bytes): Second half of derived key

        Returns:
            tuple[bytes, bytes]: Two encrypted halves
        """

        # Use derived_half_2 as AES key
        aes_enc = AesEcbEncrypter(derived_half_2)
        aes_enc.AutoPad(False)

        # Encrypt the first half: priv_key[0...15] xor derived_half_1[0...15]
        encrypted_half_1 = aes_enc.Encrypt(BytesUtils.Xor(priv_key_bytes[:16], derived_half_1[:16]))
        # Encrypt the second half: priv_key[16...31] xor derived_half_1[16...31]
        encrypted_half_2 = aes_enc.Encrypt(BytesUtils.Xor(priv_key_bytes[16:], derived_half_1[16:]))

        return encrypted_half_1, encrypted_half_2


class Bip38NoEcDecrypter:
    """
    BIP38 decrypter class.
    It decrypts a private key using the algorithm specified in BIP38 without EC multiplication.
    """

    @staticmethod
    def Decrypt(priv_key_enc: str,
                passphrase: str) -> Tuple[bytes, Bip38PubKeyModes]:
        """
        Decrypt the specified private key.

        Args:
            priv_key_enc (str): Encrypted private key bytes
            passphrase (str)  : Passphrase

        Returns:
            tuple[bytes, Bip38PubKeyModes]: Decrypted private key (index 0), public key mode (index 1)

        Raises:
            Base58ChecksumError: If base58 checksum is not valid
            ValueError: If the encrypted key is not valid
        """

        # Decode private key
        priv_key_enc_bytes = Base58Decoder.CheckDecode(priv_key_enc)
        # Check length
        if len(priv_key_enc_bytes) != Bip38NoEcConst.ENC_KEY_BYTE_LEN:
            raise ValueError(f"Invalid encrypted key length ({len(priv_key_enc_bytes)})")

        # Get all the parts back
        prefix = priv_key_enc_bytes[:2]
        flagbyte = IntegerUtils.ToBytes(priv_key_enc_bytes[2])
        address_hash = priv_key_enc_bytes[3:7]
        encrypted_half_1 = priv_key_enc_bytes[7:23]
        encrypted_half_2 = priv_key_enc_bytes[23:]

        # Check prefix and flagbyte
        if prefix != Bip38NoEcConst.ENC_KEY_PREFIX:
            raise ValueError(f"Invalid prefix ({BytesUtils.ToHexString(prefix)})")
        if flagbyte not in (Bip38NoEcConst.FLAGBYTE_COMPRESSED, Bip38NoEcConst.FLAGBYTE_UNCOMPRESSED):
            raise ValueError(f"Invalid flagbyte ({BytesUtils.ToHexString(flagbyte)})")

        # Derive key halves from the passphrase and address hash
        derived_half_1, derived_half_2 = _Bip38NoEcUtils.DeriveKeyHalves(passphrase, address_hash)
        # Get the private key back by decrypting
        priv_key_bytes = Bip38NoEcDecrypter.__DecryptAndGetPrivKey(encrypted_half_1,
                                                                   encrypted_half_2,
                                                                   derived_half_1,
                                                                   derived_half_2)

        # Get public key mode
        pub_key_mode = (Bip38PubKeyModes.COMPRESSED
                        if flagbyte == Bip38NoEcConst.FLAGBYTE_COMPRESSED
                        else Bip38PubKeyModes.UNCOMPRESSED)

        # Verify the address hash
        address_hash_got = _Bip38NoEcUtils.AddressHash(priv_key_bytes, pub_key_mode)
        if address_hash != address_hash_got:
            raise ValueError(
                f"Invalid address hash (expected: {BytesUtils.ToHexString(address_hash)}, "
                f"got: {BytesUtils.ToHexString(address_hash_got)})"
            )

        return priv_key_bytes, pub_key_mode

    @staticmethod
    def __DecryptAndGetPrivKey(encrypted_half_1: bytes,
                               encrypted_half_2: bytes,
                               derived_half_1: bytes,
                               derived_half_2: bytes) -> bytes:
        """
        Decrypt and get back private key.

        Args:
            encrypted_half_1 (bytes): First encrypted half
            encrypted_half_2 (bytes): Second encrypted half
            derived_half_1 (bytes)  : First half of derived key
            derived_half_2 (bytes)  : Second half of derived key

        Returns:
            bytes: Decrypted private key
        """

        # Use derived_half_2 as AES key
        aes_dec = AesEcbDecrypter(derived_half_2)
        aes_dec.AutoUnPad(False)

        # Decrypt using AES256Decrypt
        decrypted_half_1 = aes_dec.Decrypt(encrypted_half_1)
        decrypted_half_2 = aes_dec.Decrypt(encrypted_half_2)

        # Get the private key back by XORing bytes
        return BytesUtils.Xor(decrypted_half_1 + decrypted_half_2, derived_half_1)
